import logging
import json
import queue
import time
import os
from typing import Optional
import sys
import importlib

import numpy as np
import cv2
import rclpy
from rclpy import Parameter
from rclpy.node import Node
import std_msgs.msg
import sensor_msgs.msg

from wrapyfi.connect.listeners import Listener, Listeners, ListenerWatchDog
from wrapyfi.middlewares.ros2 import ROS2Middleware
from wrapyfi.encoders import JsonDecodeHook


WAIT = {True: None, False: 0}
WATCHDOG_POLL_REPEAT = None
QUEUE_SIZE = int(os.environ.get("WRAPYFI_ROS2_QUEUE_SIZE", 5))


class ROS2Listener(Listener, Node):

    def __init__(self, name: str, in_topic: str, should_wait: bool = True,
                 queue_size: int = QUEUE_SIZE, ros2_kwargs: Optional[dict] = None, **kwargs):
        """
        Initialize the subscriber.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param queue_size: int: Size of the queue for the subscriber. Default is 5
        :param ros2_kwargs: dict: Additional kwargs for the ROS 2 middleware
        :param kwargs: dict: Additional kwargs for the subscriber
        """
        carrier = "tcp"
        if "carrier" in kwargs and kwargs["carrier"] not in ["", None]:
            logging.warning("[ROS 2] ROS 2 currently does not support explicit carrier setting for PUB/SUB pattern. Using TCP.")
        if "carrier" in kwargs:
            del kwargs["carrier"]

        ROS2Middleware.activate(**ros2_kwargs or {})
        Listener.__init__(self, name, in_topic, carrier=carrier, should_wait=should_wait, **kwargs)
        Node.__init__(self, name + str(hex(id(self))), allow_undeclared_parameters=True)

        self.queue_size = queue_size

    def close(self):
        """
        Close the subscriber.
        """
        if hasattr(self, "_subscriber") and self._subscriber:
            if self._subscriber is not None:
                self.destroy_node()

    def __del__(self):
        self.close()


@Listeners.register("NativeObject", "ros2")
class ROS2NativeObjectListener(ROS2Listener):

    def __init__(self, name: str, in_topic: str, should_wait: bool = True, queue_size: int = QUEUE_SIZE,
                 deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        The NativeObject listener using the ROS 2 String message assuming the data is serialized as a JSON string.
        Deserializes the data (including plugins) using the decoder and parses it to a native object.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param queue_size: int: Size of the queue for the subscriber. Default is 5
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, in_topic, should_wait=should_wait, queue_size=queue_size, **kwargs)

        self._subscriber = self._queue = None

        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        ListenerWatchDog().add_listener(self)

    def establish(self):
        """
        Establish the subscriber.
        """
        self._queue = queue.Queue(maxsize=0 if self.queue_size is None or self.queue_size <= 0 else self.queue_size)
        self._subscriber = self.create_subscription(std_msgs.msg.String, self.in_topic, callback=self._message_callback, qos_profile=self.queue_size)
        self.established = True

    def listen(self):
        """
        Listen for a message.

        :return: Any: The received message as a native python object
        """
        if not self.established:
            self.establish()
        try:
            rclpy.spin_once(self, timeout_sec=WAIT[self.should_wait])
            obj_str = self._queue.get(block=self.should_wait)
            return json.loads(obj_str, object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs)
        except queue.Empty:
            return None

    def _message_callback(self, msg):
        """
        Callback for the subscriber.

        :param msg: std_msgs.msg.String: The received message
        """
        try:
            self._queue.put(msg.data, block=False)
        except queue.Full:
            logging.warning(f"[ROS 2] Discarding data because listener queue is full: {self.in_topic}")


@Listeners.register("Image", "ros2")
class ROS2ImageListener(ROS2Listener):

    def __init__(self, name: str, in_topic: str, should_wait: bool = True, queue_size: int = QUEUE_SIZE,
                 width: int = -1, height: int = -1, rgb: bool = True, fp: bool = False, jpg: bool = False, **kwargs):
        """
        The Image listener using the ROS 2 Image message parsed to a numpy array.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param queue_size: int: Size of the queue for the subscriber. Default is 5
        :param width: int: Width of the image. Default is -1 (use the width of the received image)
        :param height: int: Height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be decompressed from JPG. Default is False
        """
        super().__init__(name, in_topic, should_wait=should_wait, queue_size=queue_size, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        if self.fp:
            self._encoding = '32FC3' if self.rgb else '32FC1'
            self._type = np.float32
        else:
            self._encoding = 'bgr8' if self.rgb else 'mono8'
            self._type = np.uint8
        if self.jpg:
            self._encoding = 'jpeg'
            self._type = np.uint8

        self._pixel_bytes = (3 if self.rgb else 1) * np.dtype(self._type).itemsize

        self._subscriber = self._queue = None

        ListenerWatchDog().add_listener(self)

    def establish(self):
        """
        Establish the subscriber
        """
        self._queue = queue.Queue(maxsize=0 if self.queue_size is None or self.queue_size <= 0 else self.queue_size)
        if self.jpg:
            self._subscriber = self.create_subscription(sensor_msgs.msg.CompressedImage, self.in_topic, callback=self._message_callback, qos_profile=self.queue_size)
        else:
            self._subscriber = self.create_subscription(sensor_msgs.msg.Image, self.in_topic, callback=self._message_callback, qos_profile=self.queue_size)
        self.established = True

    def listen(self):
        """
        Listen for a message.

        :return: np.ndarray: The received message as a numpy array formatted as a cv2 image np.ndarray[img_height, img_width, channels]
        """
        if not self.established:
            self.establish()
        try:
            rclpy.spin_once(self, timeout_sec=WAIT[self.should_wait])
            if self.jpg:
                format, data = self._queue.get(block=self.should_wait)
                if format != 'jpeg':
                    raise ValueError(f"Unsupported image format: {format}")
                if self.rgb:
                    img = cv2.imdecode(np.frombuffer(data, np.uint8), cv2.IMREAD_COLOR)
                else:
                    img = cv2.imdecode(np.frombuffer(data, np.uint8), cv2.IMREAD_GRAYSCALE)
            else:
                height, width, encoding, is_bigendian, data = self._queue.get(block=self.should_wait)
                if encoding != self._encoding:
                    raise ValueError("Incorrect encoding for listener")
                if 0 < self.width != width or 0 < self.height != height or len(data) != height * width * self._pixel_bytes:
                    raise ValueError("Incorrect image shape for listener")
                img = np.frombuffer(data, dtype=np.dtype(self._type).newbyteorder('>' if is_bigendian else '<')).reshape((height, width, -1))
                if img.shape[2] == 1:
                    img = img.squeeze(axis=2)
            return img
        except queue.Empty:
            return None

    def _message_callback(self, msg):
        """
        Callback for the subscriber.

        :param msg: sensor_msgs.msg.Image: The received message
        """
        try:
            if self.jpg:
                self._queue.put((msg.format, msg.data), block=False)
            else:
                self._queue.put((msg.height, msg.width, msg.encoding, msg.is_bigendian, msg.data), block=False)
        except queue.Full:
            logging.warning(f"[ROS 2] Discarding data because listener queue is full: {self.in_topic}")


@Listeners.register("AudioChunk", "ros2")
class ROS2AudioChunkListener(ROS2Listener):

    def __init__(self, name: str, in_topic: str, should_wait: bool = True,
                 queue_size: int = QUEUE_SIZE, channels: int = 1, rate: int = 44100, chunk: int = -1, **kwargs):
        """
        The AudioChunk listener using the ROS 2 Audio message parsed to a numpy array.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param queue_size: int: Size of the queue for the subscriber. Default is 5
        :param channels: int: Number of channels in the audio. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: Number of samples in the audio chunk. Default is -1 (use the chunk size of the received audio)
        """
        super().__init__(name, in_topic, should_wait=should_wait, queue_size=queue_size, **kwargs)
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

        self._subscriber = self._queue = None

        ListenerWatchDog().add_listener(self)

    def establish(self):
        """
        Establish the subscriber.
        """
        try:
            from wrapyfi_ros2_interfaces.msg import ROS2AudioMessage
        except ImportError:
            import wrapyfi
            logging.error("[ROS 2] Could not import ROS2AudioMessage. "
                          "Make sure the ROS 2 services in wrapyfi_extensions/wrapyfi_ros2_interfaces are compiled. "
                          "Refer to the documentation for more information: \n" +
                          wrapyfi.__doc__ + "ros2_interfaces_lnk.html")
            sys.exit(1)
        self._queue = queue.Queue(maxsize=0 if self.queue_size is None or self.queue_size <= 0 else self.queue_size)
        self._subscriber = self.create_subscription(ROS2AudioMessage, self.in_topic, callback=self._message_callback, qos_profile=self.queue_size)
        self.established = True

    def listen(self):
        """
        Listen for a message.

        :return: Tuple[np.ndarray, int]: The received message as a numpy array formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        if not self.established:
            self.establish()
        try:
            rclpy.spin_once(self, timeout_sec=WAIT[self.should_wait])
            chunk, channels, rate, encoding, is_bigendian, data = self._queue.get(block=self.should_wait)
            if 0 < self.rate != rate:
                raise ValueError("Incorrect audio rate for publisher")
            if encoding not in ['S16LE', 'S16BE']:
                raise ValueError("Incorrect encoding for listener")
            if 0 < self.chunk != chunk or self.channels != channels or len(data) != chunk * channels * 4:
                raise ValueError("Incorrect audio shape for listener")
            aud = np.frombuffer(data, dtype=np.dtype(np.float32).newbyteorder('>' if is_bigendian else '<')).reshape((chunk, channels))
            # aud = aud / 32767.0
            return aud, rate
        except queue.Empty:
            return None, self.rate

    def _message_callback(self, msg):
        """
        Callback for the subscriber.

        :param msg: wrapyfi_ros2_interfaces.msg.ROS2AudioMessage: The received message
        """
        try:
            self._queue.put((msg.chunk_size, msg.channels, msg.sample_rate, msg.encoding, msg.is_bigendian, msg.data), block=False)
        except queue.Full:
            logging.warning(f"[ROS 2] Discarding data because listener queue is full: {self.in_topic}")


@Listeners.register("Properties", "ros2")
class ROS2PropertiesListener(ROS2Listener):
    def __init__(self, name, in_topic, **kwargs):
        super().__init__(name, in_topic, **kwargs)
        raise NotImplementedError


@Listeners.register("ROS2Message", "ros2")
class ROS2MessageListener(ROS2Listener):

    def __init__(self, name: str, in_topic: str, should_wait: bool = True, queue_size: int = QUEUE_SIZE, **kwargs):
        """
        The ROS2MessageListener using the ROS 2 message type inferred from the message type. Supports standard ROS 2 msgs.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param queue_size: int: Size of the queue for the subscriber. Default is 5
        """
        super().__init__(name, in_topic, should_wait=should_wait, queue_size=queue_size, **kwargs)
        self._queue = queue.Queue(maxsize=0 if self.queue_size is None or self.queue_size <= 0 else self.queue_size)

    def get_topic_type(self, topic_name):
        """
        Get the type of a specific topic.

        :param topic_name: str: Name of the topic to get its type
        :return: str or None: The topic type as a string, or None if the topic does not exist
        """
        topic_names_and_types = self.get_topic_names_and_types()
        for name, types in topic_names_and_types:
            if name == topic_name:
                return types[0]
        return None

    def establish(self):
        """
        Establish the subscriber.
        """
        while True:
            topic_type_str = self.get_topic_type(self.in_topic)
            if not self.should_wait:
                break
            if topic_type_str:
                break
        if not topic_type_str:
            return None

        module_name, class_name = topic_type_str.rsplit('/', 1)
        module_name = module_name.replace('/', '.')
        MessageType = getattr(importlib.import_module(module_name), class_name)

        self._subscriber = self.create_subscription(MessageType, self.in_topic, callback=self._message_callback, qos_profile=self.queue_size)
        self.established = True

    def listen(self):
        """
        Listen for a message.

        :return: ROS2Message: The received message as a ROS 2 message object
        """
        if not self.established:
            self.establish()
        try:
            rclpy.spin_once(self, timeout_sec=WAIT[self.should_wait])
            data = self._queue.get(block=self.should_wait)

            return data
        except queue.Empty:
            return None

    def _message_callback(self, msg):
        """
        Callback for the subscriber.

        :param msg: ROS2Message: The received message
        """
        try:
            self._queue.put(msg, block=False)
        except queue.Full:
            logging.warning(f"[ROS 2] Discarding data because listener queue is full: {self.in_topic}")



import argparse

import cv2

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR
from wrapyfi.config.manager import ConfigManager


class ExperimentController(MiddlewareCommunicator):
    WEBCAM_HEIGHT = 240
    WEBCAM_WIDTH = 320
    PEPPER_CAM_HEIGHT = 480
    PEPPER_CAM_WIDTH = 640
    ICUB_CAM_HEIGHT = 240
    ICUB_CAM_WIDTH = 320
    ESR_CAM_HEIGHT = 240
    ESR_CAM_WIDTH = 320

    def __init__(self, **kwargs):
        super(ExperimentController, self).__init__()

    @MiddlewareCommunicator.register("NativeObject", "$_mware", "ExperimentController",
                                     "/control_interface/facial_expressions_esr9", should_wait=False)
    def listen_facial_expressions_esr9(self, _mware=DEFAULT_COMMUNICATOR):
        return None,

    @MiddlewareCommunicator.register("NativeObject", "$_mware", "ExperimentController",
                                     "/control_interface/facial_expressions_icub", should_wait=False)
    def publish_facial_expressions_icub(self, obj, _mware="yarp"):
        return obj,

    @MiddlewareCommunicator.register("NativeObject", "$_mware", "ExperimentController",
                                     "/control_interface/facial_expressions_pepper",
                                     carrier="tcp", should_wait=False)
    def publish_facial_expressions_pepper(self, obj, _mware="ros"):
        return obj,

    @MiddlewareCommunicator.register("Image", "$_mware", "ExperimentController", "$_topic",
                                     width="$_width", height="$_height", rgb=True, fp=False, should_wait=False,
                                     jpg=True)
    def listen_image_webcam(self, _mware=DEFAULT_COMMUNICATOR, _width=WEBCAM_WIDTH, _height=WEBCAM_HEIGHT,
                            _topic="/control_interface/image_webcam"):  # dynamic topic according to corresponding API
        return None,

    @MiddlewareCommunicator.register("Image", "$_mware", "ExperimentController", "$_topic",
                                     width="$_width", height="$_height", rgb=True, fp=False, should_wait=False)
    def listen_image_pepper_cam(self, _mware="ros", _width=PEPPER_CAM_WIDTH, _height=PEPPER_CAM_HEIGHT,
                                _topic="/pepper/camera/front/camera/image_raw"):
        return None,

    @MiddlewareCommunicator.register("Image", "$_mware", "ExperimentController", "$_topic",
                                     width="$_width", height="$_height", rgb=True, fp=False, should_wait=False)
    def listen_image_icub_cam(self, _mware="yarp", _width=ICUB_CAM_WIDTH, _height=ICUB_CAM_HEIGHT,
                              _topic="/icub/cam/right"):
        return None,

    @MiddlewareCommunicator.register("Image", "$_mware", "ExperimentController",
                                     "/control_interface/image_webcam",
                                     width="$_width", height="$_height", rgb=True, fp=False, should_wait=False)
    def forward_image_webcam(self, img, _mware=DEFAULT_COMMUNICATOR, _width=WEBCAM_WIDTH,
                             _height=WEBCAM_HEIGHT):
        return img,

    @MiddlewareCommunicator.register("Image", "$_mware", "ExperimentController",
                                     "/control_interface/image_pepper_cam",
                                     width="$_width", height="$_height", rgb=True, fp=False, jpg=True,
                                     should_wait=False)
    def forward_image_pepper_cam(self, img, _mware=DEFAULT_COMMUNICATOR, _width=PEPPER_CAM_WIDTH,
                                 _height=PEPPER_CAM_HEIGHT):
        return img,

    @MiddlewareCommunicator.register("Image", "$_mware", "ExperimentController",
                                     "/control_interface/image_icub_cam",
                                     width="$_width", height="$_height", rgb=True, fp=False, jpg=True,
                                     should_wait=False)
    def forward_image_icub_cam(self, img, _mware=DEFAULT_COMMUNICATOR, _width=ICUB_CAM_WIDTH,
                               _height=ICUB_CAM_HEIGHT):
        # convert to bgr
        if img is not None:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return img,

    @MiddlewareCommunicator.register("Image", "$_mware", "ExperimentController",
                                     "/control_interface/image_esr9",
                                     width="$_width", height="$_height", rgb=True, fp=False, should_wait=False, jpg=True)
    def publish_esr9_cam(self, img, _mware=DEFAULT_COMMUNICATOR, _width=ESR_CAM_WIDTH, _height=ESR_CAM_HEIGHT):
        # modify size of image
        if img is not None:
            img = cv2.resize(img, (self.ESR_CAM_WIDTH, self.ESR_CAM_HEIGHT))
        return img,


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--wrapyfi_cfg",
                        help="File to load Wrapyfi configs for running instance. "
                             "Choose one of the configs available in "
                             "./wrapyfi_configs/affective_signaling_multirobot "
                             "for each running instance. All configs with a prefix of COMP must "
                             "have corresponding instances running. OPT prefixed configs execute "
                             "scripts on a machine connected to either robot (Pepper/iCub) or both, "
                             "and at least one must run in addition to COMP.",
                        type=str)
    parser.add_argument("--cam_source",
                        help="The camera input source being either from a "
                             "webcam, Pepper, or iCub. Note that this must be similar for all running "
                             "instances even when the two robots are set to display the expressions.",
                        type=str, default="webcam", choices=["webcam", "pepper", "icub"])
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    if args.wrapyfi_cfg:
        print("Wrapyfi config loading from ", args.wrapyfi_cfg)
        ConfigManager(args.wrapyfi_cfg)

    ec = ExperimentController()
    image_cam = None
    while True:
        if args.cam_source == "webcam":
            image_webcam, = ec.listen_image_webcam()
            image_webcam_linked, = ec.forward_image_webcam(image_webcam)
            if image_webcam is None:
                image_webcam = image_webcam_linked
            image_cam = image_webcam
            if image_webcam is not None:
                cv2.imshow("Webcam image", image_webcam)
                cv2.waitKey(1)
        if args.cam_source == "pepper":
            image_pepper_cam, = ec.listen_image_pepper_cam()
            image_pepper_cam_linked, = ec.forward_image_pepper_cam(image_pepper_cam)
            if image_pepper_cam is None:
                image_pepper_cam = image_pepper_cam_linked
            image_cam = image_pepper_cam
            if image_pepper_cam is not None:
                cv2.imshow("Pepper image", image_pepper_cam)
                cv2.waitKey(1)
        if args.cam_source == "icub":
            image_icub_cam, = ec.listen_image_icub_cam()
            image_icub_cam_linked, = ec.forward_image_icub_cam(image_icub_cam)
            if image_icub_cam is None:
                image_icub_cam = image_icub_cam_linked
            image_cam = image_icub_cam
            if image_icub_cam_linked is not None:
                cv2.imshow("iCub image", image_icub_cam_linked)
                cv2.waitKey(1)

        image_esr, = ec.publish_esr9_cam(image_cam)
        facial_expression, = ec.listen_facial_expressions_esr9()
        if facial_expression is not None:
            ec.publish_facial_expressions_icub(facial_expression)
            ec.publish_facial_expressions_pepper(facial_expression)




"""
Encoder and Decoder for PIL Image Data via Wrapyfi.

This script provides mechanisms to encode and decode PIL Image data using Wrapyfi.
It utilizes base64 encoding to convert binary data into ASCII strings, and also handles non-ASCII encodings for certain image formats.

The script contains a class, `PILImage`, registered as a plugin to manage the
conversion of PIL Image data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - PIL (Pillow): A Python Imaging Library that adds image processing capabilities to your Python interpreter (refer to https://pillow.readthedocs.io/en/stable/installation.html for installation instructions)
        Note: If PIL is not available, HAVE_PIL will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install Pillow``  # Basic installation of Pillow (PIL Fork)
"""

import io
import base64

from wrapyfi.utils import *

try:
    from PIL import Image
    HAVE_PIL = True
except ImportError:
    HAVE_PIL = False


@PluginRegistrar.register(types=None if not HAVE_PIL else Image.Image.__mro__[:-1])
class PILImage(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the PILImage plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode PIL Image data into a base64 ASCII string.

        :param obj: Image.Image: The PIL Image data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name and encoded data string, with optional image size and mode for raw data
        """
        if obj.format is None:
            obj_data = base64.b64encode(obj.tobytes()).decode('ascii')
            return True, dict(__wrapyfi__=(str(self.__class__.__name__), obj_data, obj.size, obj.mode))
        else:
            with io.BytesIO() as memfile:
                obj.save(memfile, format=obj.format)
                obj_data = memfile.getvalue().decode('latin1')
            return True, dict(__wrapyfi__=(str(self.__class__.__name__), obj_data))

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into PIL Image data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string and optionally image size and mode for raw data
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, Image.Image]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - Image.Image: The decoded PIL Image data
        """
        if len(obj_full) == 4:
            with io.BytesIO(obj_full[1].encode('ascii')) as memfile:
                return True, Image.frombytes(obj_full[3], obj_full[2], memfile.getvalue(), "raw")
        else:
            with io.BytesIO(obj_full[1].encode('latin1')) as memfile:
                memfile.seek(0)
                return True, Image.open(memfile).copy()




import logging

from wrapyfi.connect.listeners import Listener, Listeners, ListenerWatchDog


@Listeners.register("MMO", "fallback")
class FallbackListener(Listener):

    def __init__(self, name: str, in_topic: str, carrier: str = "tcp",
                 should_wait: bool = True, missing_middleware_object: str = "", **kwargs):
        logging.warning(f"Fallback listener employed due to missing middleware or object type: "
                        f"{missing_middleware_object}")
        Listener.__init__(self, name, in_topic, carrier=carrier, should_wait=should_wait, **kwargs)
        self.missing_middleware_object = missing_middleware_object

    def establish(self, repeats: int = -1, **kwargs):
        return None

    def listen(self):
        return None

    def close(self):
        return None


"""
Encoder and Decoder for pandas Series/Dataframes Data via Wrapyfi.

This script provides mechanisms to encode and decode pandas data using Wrapyfi.
It utilizes base64 encoding to convert binary data into ASCII strings.

The script contains a class, `PandasData`, registered as a plugin to manage the
conversion of pandas data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - pandas: A data structures library for data analysis, time series, and statistics (refer to https://pandas.pydata.org/pandas-docs/stable/getting_started/install.html for installation instructions)
        Note: If pandas is not available, HAVE_PANDAS will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install "pandas<2.0"``  # Basic installation of pandas
"""

import io
import base64

import numpy as np

from wrapyfi.utils import *

try:
    import pandas
    HAVE_PANDAS = True
except ImportError:
    HAVE_PANDAS = False


@PluginRegistrar.register(types=None if not HAVE_PANDAS else pandas.DataFrame.__mro__[:-1] + pandas.Series.__mro__[:-1])
class PandasData(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the PandasData plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode pandas data into a base64 ASCII string.

        :param obj: Union[pandas.DataFrame, pandas.Series]: The pandas data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, encoded data string, and object type
        """
        with io.BytesIO() as memfile:
            if isinstance(obj, pandas.DataFrame):
                obj_type = 'DataFrame'
                obj.to_json(memfile, orient="split")
            elif isinstance(obj, pandas.Series):
                obj_type = 'Series'
                obj.to_frame().to_json(memfile, orient="split")
            obj_data = base64.b64encode(memfile.getvalue()).decode('ascii')
        return True, dict(__wrapyfi__=(str(self.__class__.__name__), obj_data, obj_type))

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into pandas data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string and object type
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, Union[pandas.DataFrame, pandas.Series]]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - Union[pandas.DataFrame, pandas.Series]: The decoded pandas data
        """
        with io.BytesIO(base64.b64decode(obj_full[1].encode('ascii'))) as memfile:
            obj = pandas.read_json(memfile, orient="split")
            obj_type = obj_full[2]
            if obj_type == 'Series':
                obj = obj.iloc[:, 0]
        return True, obj




"""
Encoder and Decoder for XArray DataArray/Dataset Data via Wrapyfi.

This script provides mechanisms to encode and decode XArray Data using Wrapyfi.
It utilizes base64 encoding and JSON serialization.

The script contains a class, `XArrayData`, registered as a plugin to manage the
conversion of XArray Data (DataArray and Dataset) (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - pandas: A data structures library for data analysis, time series, and statistics (refer to https://pandas.pydata.org/pandas-docs/stable/getting_started/install.html for installation instructions)
    - XArray: N-D labeled arrays and datasets in Python (refer to http://xarray.pydata.org/en/stable/getting-started-guide/installing.html for installation instructions)
        Note: If XArray is not available, HAVE_XARRAY will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install "pandas<2.0" xarray``  # Basic installation of XArray
"""

import json
import base64
from datetime import datetime

import numpy as np

from wrapyfi.utils import *

try:
    import xarray as xr
    HAVE_XARRAY = True
except ImportError:
    HAVE_XARRAY = False


@PluginRegistrar.register(types=None if not HAVE_XARRAY else xr.DataArray.__mro__[:-1] + xr.Dataset.__mro__[:-1])
class XArrayData(Plugin):
    def __init__(self, **kwargs):
        """
        Initialize the XArrayData plugin.
        """
        pass

    def encode(self, obj, *args, **kwargs):
        """
        Encode XArray Data using JSON and base64.

        :param obj: Union[xr.DataArray, xr.Dataset]: The XArray Data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, encoded data string, data type, and object name
        """
        obj_type = 'Dataset'
        obj_name = None
        if isinstance(obj, xr.DataArray):
            obj_dict = obj.to_dataset().to_dict()
            obj_type = 'DataArray'
            obj_name = obj.name
        elif isinstance(obj, xr.Dataset):
            obj_dict = obj.to_dict()

        def traverse_and_convert(obj):
            if isinstance(obj, list):
                return [traverse_and_convert(item) for item in obj]
            elif isinstance(obj, dict):
                return {key: traverse_and_convert(value) for key, value in obj.items()}
            elif isinstance(obj, np.datetime64):
                return {'data': str(obj), 'dtype': str(obj.dtype)}
            elif isinstance(obj, datetime):
                return {'data': obj.isoformat(), 'dtype': 'datetime'}
            else:
                return obj

        converted_obj_dict = traverse_and_convert(obj_dict)
        obj_json = json.dumps(converted_obj_dict)
        obj_data = base64.b64encode(obj_json.encode('ascii')).decode('ascii')

        return True, dict(__wrapyfi__=(str(self.__class__.__name__), obj_data, obj_type, obj_name))

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a JSON and base64 encoded string back into XArray Data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string, data type, and object name
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, Union[xr.DataArray, xr.Dataset]]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - Union[xr.DataArray, xr.Dataset]: The decoded XArray Data
        """
        obj_data = base64.b64decode(obj_full[1].encode('ascii')).decode('ascii')
        xarray_type = obj_full[2]
        xarray_name = obj_full[3]

        def traverse_and_reconvert(obj):
            if isinstance(obj, list):
                return [traverse_and_reconvert(item) for item in obj]
            elif isinstance(obj, dict):
                if 'data' in obj and 'dtype' in obj:
                    if obj['dtype'] == 'datetime':
                        return datetime.fromisoformat(obj['data'])
                    else:
                        return np.datetime64(obj['data'], obj['dtype'])
                return {key: traverse_and_reconvert(value) for key, value in obj.items()}
            else:
                return obj

        obj_dict = json.loads(obj_data)
        reconverted_obj_dict = traverse_and_reconvert(obj_dict)

        if xarray_type == 'DataArray':
            dataset = xr.Dataset.from_dict(reconverted_obj_dict)
            data_array = dataset[xarray_name]
            return True, data_array
        else:
            return True, xr.Dataset.from_dict(reconverted_obj_dict)


import os
import re


def get_project_info_from_setup():
    # when Wrapyfi is not installed
    try:
        curr_dir = os.path.dirname(__file__)
        setup_path = os.path.join(curr_dir, "..", "setup.py")
        with open(setup_path, 'r') as f:
            content = f.read()
    except FileNotFoundError:
        return {}
    name_match = re.search(r"name\s*=\s*['\"]([^'\"]*)['\"]", content)
    version_match = re.search(r"version\s*=\s*['\"]([^'\"]*)['\"]", content)
    url_match = re.search(r"url\s*=\s*['\"]([^'\"]*)['\"]", content)
    doc_match = re.search(r"'Documentation':\s*['\"]([^'\"]*)['\"]", content)
    
    if not name_match or not version_match or not url_match:
        # raise RuntimeError("Unable to find name, version, or url string.")
        return {}
        
    return {
        'name': name_match.group(1),
        'version': version_match.group(1),
        'url': url_match.group(1),
        'doc': None if not doc_match else doc_match.group(1)
    }


# extract project info
project_info = get_project_info_from_setup()

__version__ = project_info.get('version', None)
__url__ = project_info.get('url', None)
__doc__ = project_info.get('doc', None)
name = project_info.get('name', 'wrapyfi')

if __version__ is None or __url__ is None or __doc__ is None:
    try:
        from importlib import metadata
        mdata = metadata.metadata(__name__)
        __version__ = metadata.version(__name__)
        __url__ = mdata["Home-page"]
        # when installed with PyPi
        if __url__ is None:
            for url_extract in mdata.get_all("Project-URL"):
                __url__ = url_extract.split(", ")[1] if url_extract.split(", ")[0] == "Homepage" else __url__
        if __doc__ is None:
            for url_extract in mdata.get_all("Project-URL"):
                __doc__ = url_extract.split(", ")[1] if url_extract.split(", ")[0] == "Documentation" else __doc__
    except ImportError:
        try:
            # when Python < 3.8 and setuptools/pip have not been updated
            import pkg_resources
            mdata = pkg_resources.get_distribution(__name__).metadata
            __version__ = pkg_resources.require(__name__)[0].version
            __url__ = mdata["Home-page"]
            # when installed with PyPi
            if __url__ is None:
                for url_extract in mdata.get_all("Project-URL"):
                    __url__ = url_extract.split(", ")[1] if url_extract.split(", ")[0] == "Homepage" else __url__
            if __doc__ is None:
                for url_extract in mdata.get_all("Project-URL"):
                    __doc__ = url_extract.split(", ")[1] if url_extract.split(", ")[0] == "Documentation" else __doc__
        except pkg_resources.DistributionNotFound:
            __version__ = "unknown_version"
            __url__ = "unknown_url"
            __doc__ = "unknown_url"
    except Exception:
        __version__ = "unknown_version"
        __url__ = "unknown_url"
        __doc__ = "unknown_url"

from wrapyfi.utils import PluginRegistrar

PluginRegistrar.scan()

import logging
logging.getLogger().setLevel(logging.INFO)



import logging
import json
import time
import base64
import io
from typing import Optional, Literal

import numpy as np
import cv2
import yarp

from wrapyfi.connect.listeners import Listener, Listeners, ListenerWatchDog
from wrapyfi.middlewares.yarp import YarpMiddleware
from wrapyfi.encoders import JsonDecodeHook


WATCHDOG_POLL_REPEAT = None


class YarpListener(Listener):

    def __init__(self, name: str, in_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp", should_wait: bool = True,
                 persistent: bool = True, yarp_kwargs: Optional[dict] = None, **kwargs):
        """
        Initialize the subscriber.

        :param name: str: Name of the publisher
        :param in_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param persistent: bool: Whether the subscriber port should remain connected after closure. Default is True
        :param yarp_kwargs: dict: Additional kwargs for  the Yarp middleware
        :param kwargs: dict: Additional kwargs for the subscriber
        """
        super().__init__(name, in_topic, carrier=carrier, should_wait=should_wait, **kwargs)
        self.style = yarp.ContactStyle()
        self.style.persistent = persistent
        self.style.carrier = self.carrier

        YarpMiddleware.activate(**yarp_kwargs or {})

    def await_connection(self, in_topic: Optional[str] = None, repeats: Optional[int] = None):
        """
        Wait for the publisher to connect to the subscriber.

        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param repeats: int: Number of times to check for the parameter. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        connected = False
        if in_topic is None:
            in_topic = self.in_topic
        logging.info(f"[YARP] Waiting for input port: {in_topic}")
        if repeats is None:
            if self.should_wait:
                repeats = -1
            else:
                repeats = 1

            while repeats > 0 or repeats <= -1:
                repeats -= 1
                connected = yarp.Network.exists(in_topic)
                if connected:
                    logging.info(f"[YARP] Connected to input port: {in_topic}")
                    break
                time.sleep(0.2)
        return connected

    def read_port(self, port):
        """
        Read the port.

        :param port: yarp.Port: Port to read from
        :return: yarp.Value: Value read from the port
        """
        while True:
            obj = port.read(shouldWait=False)
            if self.should_wait and obj is None:
                time.sleep(0.005)
            else:
                return obj

    def close(self):
        """
        Close the subscriber
        """
        if hasattr(self, "_port") and self._port:
            if self._port is not None:
                self._port.close()

    def __del__(self):
        self.close()


@Listeners.register("NativeObject", "yarp")
class YarpNativeObjectListener(YarpListener):

    def __init__(self, name: str, in_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp", should_wait: bool = True,
                 persistent: bool = True, deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        The NativeObject listener using the BufferedPortBottle string construct assuming the data is serialized as a JSON string.
        Deserializes the data (including plugins) using the decoder and parses it to a Python object.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param persistent: bool: Whether the subscriber port should remain connected after closure. Default is True
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, in_topic, carrier=carrier, should_wait=should_wait, persistent=persistent, **kwargs)
        self._port = self._netconnect = None

        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        if not self.should_wait:
            ListenerWatchDog().add_listener(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection to the publisher.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        established = self.await_connection(repeats=repeats)
        if established:
            self._port = yarp.BufferedPortBottle()
            rnd_id = str(np.random.randint(100000, size=1)[0])
            self._port.open(self.in_topic + ":in" + rnd_id)
            if self.style.persistent:
                self._netconnect = yarp.Network.connect(self.in_topic, self.in_topic + ":in" + rnd_id, self.style)
            else:
                self._netconnect = yarp.Network.connect(self.in_topic, self.in_topic + ":in" + rnd_id, self.carrier)
        return self.check_establishment(established)

    def listen(self):
        """
        Listen for a message.

        :return: Any: The received message as a native python object
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return None
        obj_port = self.read_port(self._port)
        if obj_port is not None:
            return json.loads(obj_port.get(0).asString(), object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs)
        else:
            return None


@Listeners.register("Image", "yarp")
class YarpImageListener(YarpListener):

    def __init__(self, name: str, in_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp", should_wait: bool = True,
                 persistent: bool = True, width: int = -1, height: int = -1, rgb: bool = True, fp: bool = False, jpg: bool = False, **kwargs):
        """
        The Image listener using the BufferedPortImage construct parsed to a numpy array.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param persistent: bool: Whether the subscriber port should remain connected after closure. Default is True
        :param width: int: Width of the image. Default is -1 (use the width of the received image)
        :param height: int: Height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be decompressed from JPG. Default is False
        """
        super().__init__(name, in_topic, carrier=carrier, should_wait=should_wait, persistent=persistent, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        self._port = self._type = self._netconnect = None

        if not self.should_wait:
            ListenerWatchDog().add_listener(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection to the publisher.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        established = self.await_connection(repeats=repeats)
        if established:
            if self.jpg:
                self._port = yarp.BufferedPortBottle()
            elif self.rgb:
                self._port = yarp.BufferedPortImageRgbFloat() if self.fp else yarp.BufferedPortImageRgb()
            else:
                self._port = yarp.BufferedPortImageFloat() if self.fp else yarp.BufferedPortImageMono()
            self._type = np.float32 if self.fp else np.uint8
            in_topic_connect = f"{self.in_topic}:in{np.random.randint(100000, size=1).item()}"
            self._port.open(in_topic_connect)
            if self.style.persistent:
                self._netconnect = yarp.Network.connect(self.in_topic, in_topic_connect, self.style)
            else:
                self._netconnect = yarp.Network.connect(self.in_topic, in_topic_connect, self.carrier)
        return self.check_establishment(established)

    def listen(self):
        """
        Listen for a message.

        :return: np.ndarray: The received message as a numpy array formatted as a cv2 image np.ndarray[img_height, img_width, channels]
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return None
        ret_img_msg = self.read_port(self._port)
        if ret_img_msg is None:
            return None
        if self.jpg:
            img_str = ret_img_msg.get(0).asString()
            with io.BytesIO(base64.b64decode(img_str.encode('ascii'))) as memfile:
                img_str = np.load(memfile)
            if self.rgb:
                img = cv2.imdecode(np.frombuffer(img_str, np.uint8), cv2.IMREAD_COLOR)
            else:
                img = cv2.imdecode(np.frombuffer(img_str, np.uint8), cv2.IMREAD_GRAYSCALE)
            return img
        else:
            if 0 < self.width != ret_img_msg.width() or 0 < self.height != ret_img_msg.height():
                raise ValueError("Incorrect image shape for listener")
            elif self.rgb:
                img = np.zeros((ret_img_msg.height(), ret_img_msg.width(), 3), dtype=self._type, order='C')
                img_port = yarp.ImageRgbFloat() if self.fp else yarp.ImageRgb()
            else:
                img = np.zeros((ret_img_msg.height(), ret_img_msg.width()), dtype=self._type, order='C')
                img_port = yarp.ImageFloat() if self.fp else yarp.ImageMono()
            img_port.resize(img.shape[1], img.shape[0])
            img_port.setExternal(img.data, img.shape[1], img.shape[0])
            img_port.copy(ret_img_msg)
            return img


@Listeners.register("AudioChunk", "yarp")
class YarpAudioChunkListener(YarpListener):

    def __init__(self, name: str, in_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp", should_wait: bool = True,
                 persistent: bool = True, channels: int = 1, rate: int = 44100, chunk: int = -1, **kwargs):
        """
        The AudioChunk listener using the Sound construct parsed as a numpy array.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param persistent: bool: Whether the subscriber port should remain connected after closure. Default is True
        :param channels: int: Number of channels in the audio. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: Number of samples in the audio chunk. Default is -1 (use the chunk size of the received audio)
        """
        super().__init__(name, in_topic, carrier=carrier, should_wait=should_wait, persistent=persistent, **kwargs)
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

        self._sound_msg = self._port = self._netconnect = None

        if not self.should_wait:
            ListenerWatchDog().add_listener(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection to the publisher.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        established = self.await_connection(in_topic=self.in_topic, repeats=repeats)
        if established:
            rnd_id = str(np.random.randint(100000, size=1)[0])
            self._port = yarp.Port()
            self._port.open(self.in_topic + ":in" + rnd_id)
            self._netconnect = yarp.Network.connect(self.in_topic, self.in_topic + ":in" + rnd_id, self.carrier)

            self._sound_msg = yarp.Sound()
            self._port.read(self._sound_msg)
            if self.rate == -1:
                self.rate = self._sound_msg.getFrequency()
            if self.chunk == -1:
                self.chunk = self._sound_msg.getSamples()
            if self.channels == -1:
                self.channels = self._sound_msg.getChannels()
        established = self.check_establishment(established)
        return established

    def listen(self):
        """
        Listen for a message.

        :return: Tuple[np.ndarray, int]: The received message as a numpy array formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return None
        self._port.read(self._sound_msg)
        aud = np.array([self._sound_msg.get(i) for i in range(self._sound_msg.getSamples())], dtype=np.int16)
        aud = aud.astype(np.float32) / 32767.0
        return aud, self.rate


@Listeners.register("Properties", "yarp")
class YarpPropertiesListener(YarpListener):
    def __init__(self, name, in_topic, **kwargs):
        super().__init__(name, in_topic, **kwargs)
        raise NotImplementedError



import argparse
import time
try:
    import numpy as np
    import pandas as pd
except ImportError:
    print("Install pandas and NumPy before running this script.")
try:
    import tensorflow as tf
    # avoid allocating all GPU memory assuming tf>=2.2
    gpus = tf.config.experimental.list_physical_devices('GPU')
    for gpu in gpus:
        tf.config.experimental.set_memory_growth(gpu, True)
except ImportError:
    tf = None

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


SHOULD_WAIT = True


class Benchmarker(MiddlewareCommunicator):

    @staticmethod
    def get_numpy_object(dims):
        return {"numpy": np.ones(dims)}

    @staticmethod
    def get_pandas_object(dims):
        return {"pandas": pd.DataFrame(np.ones(dims), index=None, columns=list(range(dims[-1])))}

    @staticmethod
    def get_pillow_object(dims):
        from PIL import Image
        return {"pillow": Image.fromarray(np.ones(dims, dtype=np.uint8))}

    @staticmethod
    def get_tensorflow_object(dims):
        return {"tensorflow": tf.ones(dims)}

    @staticmethod
    def get_jax_object(dims):
        import jax as jx
        return {"jax": jx.numpy.ones(dims)}

    @staticmethod
    def get_mxnet_object(dims):
        import mxnet as mx
        return {"mxnet": mx.nd.ones(dims)}

    @staticmethod
    def get_mxnet_gpu_object(dims, gpu=0):
        import mxnet as mx
        return {"mxnet_gpu": mx.nd.ones(dims, ctx=mx.gpu(gpu))}

    @staticmethod
    def get_pytorch_object(dims):
        import torch as th
        return {"pytorch": th.ones(dims)}

    @staticmethod
    def get_pytorch_gpu_object(dims, gpu=0):
        import torch as th
        return {"pytorch_gpu": th.ones(dims, device=f"cuda:{gpu}")}

    @staticmethod
    def get_paddle_object(dims):
        import paddle as pa
        return {"paddle": pa.Tensor(pa.ones(dims), place=pa.CPUPlace())}

    @staticmethod
    def get_paddle_gpu_object(dims, gpu=0):
        import paddle as pa
        return {"paddle_gpu": pa.Tensor(pa.zeros(dims), place=pa.CUDAPlace(gpu))}

    def get_all_objects(self, count, plugin_name):
        obj = {"count": count,
               "time": time.time()}
        obj.update(**getattr(self, f"get_{plugin_name}_object")((args.height, args.width,)))
        return obj

    @MiddlewareCommunicator.register("NativeObject", "yarp",
                                     "ExampleClass", "/example/get_native_objects",
                                     carrier="tcp", should_wait=SHOULD_WAIT)
    def get_yarp_native_objects(self, count, plugin_name):
        return self.get_all_objects(count, plugin_name),

    @MiddlewareCommunicator.register("NativeObject", "ros",
                                     "ExampleClass", "/example/get_native_objects",
                                     carrier="tcp", should_wait=SHOULD_WAIT)
    def get_ros_native_objects(self, count, plugin_name):
        return self.get_all_objects(count, plugin_name),

    @MiddlewareCommunicator.register("NativeObject", "ros2",
                                     "ExampleClass", "/example/get_native_objects",
                                     carrier="tcp", should_wait=SHOULD_WAIT)
    def get_ros2_native_objects(self, count, plugin_name):
        return self.get_all_objects(count, plugin_name),

    @MiddlewareCommunicator.register("NativeObject", "zeromq",
                                     "ExampleClass", "/example/get_native_objects",
                                     carrier="tcp", should_wait=SHOULD_WAIT)
    def get_zeromq_native_objects(self, count, plugin_name):
        return self.get_all_objects(count, plugin_name),


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--publish", dest="mode", action="store_const", const="publish", default="listen", help="Publish mode")
    parser.add_argument("--listen", dest="mode", action="store_const", const="listen", default="listen", help="Listen mode (default)")
    parser.add_argument("--mwares", type=str, default=list(MiddlewareCommunicator.get_communicators()),
                        choices=MiddlewareCommunicator.get_communicators(), nargs="+",
                        help="The middlewares to use for transmission")
    parser.add_argument("--plugins", type=str,
                        default=["numpy", "pandas", "tensorflow", "jax", "mxnet", "mxnet_gpu", "pytorch", "pytorch_gpu",
                                 "paddle", "paddle_gpu"], nargs="+",
                        help="The middlewares to use for transmission")
    parser.add_argument("--height", type=int, default=200, help="The tensor image height")
    parser.add_argument("--width", type=int, default=200, help="The tensor image width")
    parser.add_argument("--trials", type=int, default=2000, help="Number of trials to run per middleware")
    parser.add_argument("--skip-trials", type=int, default=0, help="Number of trials to skip before logging "
                                                               "to csv to avoid warmup time logging")
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    benchmarker = Benchmarker()
    benchmark_logger = pd.DataFrame(columns=["middleware", "plugin", "timestamp", "count", "delay"])
    benchmark_iterator = {}

    for middleware_name in args.mwares:
        benchmark_iterator[middleware_name] = getattr(benchmarker, f"get_{middleware_name}_native_objects")

    for middleware_name, method in benchmark_iterator.items():
        benchmarker.activate_communication(method, mode=args.mode)
        for plugin_name in args.plugins:
            time_acc_native_objects = []
            counter = -1
            while True:
                counter += 1
                native_objects, = method(counter, plugin_name)
                if native_objects is not None:
                    time_acc_native_objects.append(time.time() - native_objects["time"])
                    print(f"{middleware_name} :: {plugin_name} :: delay:", time_acc_native_objects[-1],
                          " Length:", len(time_acc_native_objects), " Count:", native_objects["count"])
                    if args.trials - 1 == native_objects["count"]:
                        break
                    if counter > args.skip_trials:
                        benchmark_logger = benchmark_logger.append(pd.DataFrame({"middleware": [middleware_name],
                                                                                 "plugin": [plugin_name],
                                                                                 "timestamp": [native_objects["timestamp"]],
                                                                                 "count": [native_objects["count"]],
                                                                                 "delay": [time_acc_native_objects[-1]]}),
                                                                   ignore_index=True)
                    if counter == 0:
                        if args.mode == "publish":
                            time.sleep(5)
                    else:
                        if args.mode == "publish":
                            time.sleep(0.1)

            time_acc_native_objects = pd.DataFrame(np.array(time_acc_native_objects))
            print(f"{middleware_name} :: {plugin_name} :: time statistics:", time_acc_native_objects.describe())
            time.sleep(5)
    benchmark_logger.to_csv(f"results/benchmarking_native_object_{args.mode}__{','.join(args.mwares)}__{','.join(args.plugins)}.csv", index=False)


import logging

from wrapyfi.connect.servers import Server, Servers


@Servers.register("MMO", "fallback")
class FallbackServer(Server):

    def __init__(self, name: str, out_topic: str, carrier: str = "", missing_middleware_object: str = "", **kwargs):
        logging.warning(f"Fallback server employed due to missing middleware or object type: "
                        f"{missing_middleware_object}")
        Server.__init__(self, name, out_topic, carrier=carrier, **kwargs)
        self.missing_middleware_object = missing_middleware_object

    def establish(self, repeats: int = -1, **kwargs):
        return None

    def await_request(self, *args, **kwargs):
        return args, kwargs

    def reply(self, obj):
        return obj

    def close(self):
        return None



import logging
import sys
import json
import time
import os
import base64
import io
import importlib.util
from typing import Optional, Tuple

import numpy as np
import cv2
import rospy
import std_msgs.msg
import sensor_msgs.msg

from wrapyfi.connect.publishers import Publisher, Publishers, PublisherWatchDog
from wrapyfi.middlewares.ros import ROSMiddleware
from wrapyfi.encoders import JsonEncoder


QUEUE_SIZE = int(os.environ.get("WRAPYFI_ROS_QUEUE_SIZE", 5))
WATCHDOG_POLL_REPEAT = None


class ROSPublisher(Publisher):

    def __init__(self, name: str, out_topic: str, carrier: str = "tcp", should_wait: bool = True,
                 queue_size: int = QUEUE_SIZE, ros_kwargs: Optional[dict] = None, **kwargs):
        """
        Initialize the publisher.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param queue_size: int: Queue size for the publisher. Default is 5
        :param ros_kwargs: dict: Additional kwargs for the ROS middleware
        :param kwargs: dict: Additional kwargs for the publisher
        """
        if carrier or carrier != "tcp":
            logging.warning("[ROS] ROS does not support other carriers than TCP for PUB/SUB pattern. Using TCP.")
            carrier = "tcp"
        super().__init__(name, out_topic, carrier=carrier, should_wait=should_wait, **kwargs)
        ROSMiddleware.activate(**ros_kwargs or {})

        self.queue_size = queue_size

    def await_connection(self, publisher, out_topic: Optional[str] = None, repeats: Optional[int] = None):
        """
        Wait for at least one subscriber to connect to the publisher.

        :param publisher: rospy.Publisher: Publisher to await connection to
        :param out_topic: str: Name of the output topic
        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        connected = False
        if out_topic is None:
            out_topic = self.out_topic
        logging.info(f"[ROS] Waiting for topic subscriber: {out_topic}")
        if repeats is None:
            if self.should_wait:
                repeats = -1
            else:
                repeats = 1
            while repeats > 0 or repeats <= -1:
                repeats -= 1
                connected = publisher.get_num_connections() < 1
                if connected:
                    break
                time.sleep(0.02)
        logging.info(f"[ROS] Topic subscriber connected: {out_topic}")
        return connected

    def close(self):
        """
        Close the publisher
        """
        if hasattr(self, "_publisher") and self._publisher:
            if self._publisher is not None:
                self._publisher.unregister()

    def __del__(self):
        self.close()


@Publishers.register("NativeObject", "ros")
class ROSNativeObjectPublisher(ROSPublisher):

    def __init__(self, name: str, out_topic: str, carrier: str = "tcp", should_wait: bool = True,
                 queue_size: int = QUEUE_SIZE, serializer_kwargs: Optional[dict] = None, **kwargs):
        """
        The NativeObject publisher using the ROS String message assuming a combination of python native objects.

        and numpy arrays as input. Serializes the data (including plugins) using the encoder and sends it as a string
        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param queue_size: int: Queue size for the publisher. Default is 5
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        super().__init__(name, out_topic, carrier=carrier, should_wait=should_wait, queue_size=queue_size, **kwargs)
        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

        self._publisher = None

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        self._publisher = rospy.Publisher(self.out_topic, std_msgs.msg.String, queue_size=self.queue_size)
        established = self.await_connection(self._publisher, repeats=repeats)
        return self.check_establishment(established)

    def publish(self, obj):
        """
        Publish the object to the middlware.

        :param obj: object: Object to publish
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)
        obj_str = json.dumps(obj, cls=self._plugin_encoder, **self._plugin_kwargs,
                             serializer_kwrags=self._serializer_kwargs)
        self._publisher.publish(obj_str)


@Publishers.register("Image", "ros")
class ROSImagePublisher(ROSPublisher):

    def __init__(self, name: str, out_topic: str, carrier: str = "tcp",  should_wait: bool = True, queue_size: int = QUEUE_SIZE,
                 width: int = -1, height: int = -1, rgb: bool = True, fp: bool = False, jpg: bool = False, **kwargs):
        """
        The ImagePublisher using the ROS Image message assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param queue_size: int: Queue size for the publisher. Default is 5
        :param width: int: Width of the image. Default is -1 meaning that the width is not fixed
        :param height: int: Height of the image. Default is -1 meaning that the height is not fixed
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be compressed as JPG. Default is False
        """
        super().__init__(name, out_topic, carrier=carrier, should_wait=should_wait, queue_size=queue_size, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        if self.fp:
            self._encoding = '32FC3' if self.rgb else '32FC1'
            self._type = np.float32
        else:
            self._encoding = 'bgr8' if self.rgb else 'mono8'
            self._type = np.uint8
        if self.jpg:
            self._encoding = 'jpeg'
            self._type = np.uint8

        self._publisher = None

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        if self.jpg:
            self._publisher = rospy.Publisher(self.out_topic, sensor_msgs.msg.CompressedImage, queue_size=self.queue_size)
        else:
            self._publisher = rospy.Publisher(self.out_topic, sensor_msgs.msg.Image, queue_size=self.queue_size)
        established = self.await_connection(self._publisher)
        return self.check_establishment(established)

    def publish(self, img: np.ndarray):
        """
        Publish the image to the middleware.

        :param img: np.ndarray: Image to publish formatted as a cv2 image np.ndarray[img_height, img_width, channels]
        """
        if img is None:
            return

        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)

        if 0 < self.width != img.shape[1] or 0 < self.height != img.shape[0] or \
                not ((img.ndim == 2 and not self.rgb) or (img.ndim == 3 and self.rgb and img.shape[2] == 3)):
            raise ValueError("Incorrect image shape for publisher")
        img = np.require(img, dtype=self._type, requirements='C')

        if self.jpg:
            img_msg = sensor_msgs.msg.CompressedImage()
            img_msg.header.stamp = rospy.Time.now()
            img_msg.format = "jpeg"
            img_msg.data = np.array(cv2.imencode('.jpg', img)[1]).tobytes()
        else:
            img_msg = sensor_msgs.msg.Image()
            img_msg.header.stamp = rospy.Time.now()
            img_msg.height = img.shape[0]
            img_msg.width = img.shape[1]
            img_msg.encoding = self._encoding
            img_msg.is_bigendian = img.dtype.byteorder == '>' or (img.dtype.byteorder == '=' and sys.byteorder == 'big')
            img_msg.step = img.strides[0]
            img_msg.data = img.tobytes()
        self._publisher.publish(img_msg)


@Publishers.register("AudioChunk", "ros")
class ROSAudioChunkPublisher(ROSPublisher):

    def __init__(self, name: str, out_topic: str, carrier: str = "tcp", should_wait: bool = True, queue_size: int = QUEUE_SIZE,
                 channels: int = 1, rate: int = 44100, chunk: int = -1, **kwargs):
        """
        The AudioChunkPublisher using the ROS Audio message assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param queue_size: int: Queue size for the publisher. Default is 5
        :param channels: int: Number of channels. Default is 1
        :param rate: int: Sampling rate. Default is 44100
        :param chunk: int: Chunk size. Default is -1 meaning that the chunk size is not fixed
        """
        super().__init__(name, out_topic, carrier=carrier, should_wait=should_wait, queue_size=queue_size, **kwargs)
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

        self._publisher = self._sound_msg = None

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        try:
            from wrapyfi_ros_interfaces.msg import ROSAudioMessage
        except ImportError:
            import wrapyfi
            logging.error("[ROS] Could not import ROSAudioMessage. "
                          "Make sure the ROS messages in wrapyfi_extensions/wrapyfi_ros_interfaces are compiled. "
                          "Refer to the documentation for more information: \n" +
                          wrapyfi.__doc__ + "ros_interfaces_lnk.html")
            sys.exit(1)
        self._publisher = rospy.Publisher(self.out_topic, ROSAudioMessage, queue_size=self.queue_size)
        self._sound_msg = ROSAudioMessage()
        established = self.await_connection(self._publisher)
        return self.check_establishment(established)

    def publish(self, aud: Tuple[np.ndarray, int]):
        """
        Publish the audio chunk to the middleware.

        :param aud: Tuple[np.ndarray, int]: Audio chunk to publish formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)

        aud, rate = aud
        if aud is None:
            return
        if 0 < self.rate != rate:
            raise ValueError("Incorrect audio rate for publisher")
        chunk, channels = aud.shape if len(aud.shape) > 1 else (aud.shape[0], 1)
        self.chunk = chunk if self.chunk == -1 else self.chunk
        self.channels = channels if self.channels == -1 else self.channels
        if 0 < self.chunk != chunk or 0 < self.channels != channels:
            raise ValueError("Incorrect audio shape for publisher")
        aud = np.require(aud, dtype=np.float32, requirements='C')

        aud_msg = self._sound_msg
        aud_msg.header.stamp = rospy.Time.now()
        aud_msg.chunk_size = chunk
        aud_msg.channels = channels
        aud_msg.sample_rate = rate
        aud_msg.is_bigendian = aud.dtype.byteorder == '>' or (aud.dtype.byteorder == '=' and sys.byteorder == 'big')
        aud_msg.encoding = 'S16BE' if aud_msg.is_bigendian else 'S16LE'
        aud_msg.step = aud.strides[0]
        aud_msg.data = aud.tobytes()  # (aud * 32767.0).tobytes()
        self._publisher.publish(aud_msg)


@Publishers.register("Properties", "ros")
class ROSPropertiesPublisher(ROSPublisher):
    """
    Sets rospy parameters. Behaves differently from other data types by directly setting ROS parameters.
    Note that the listener is not guaranteed to receive the updated signal, since the listener can trigger before
    property is set. The property decorated method returns accept native python objects (excluding None),
    but care should be taken when using dictionaries, since they are analogous with node namespaces:
    http://wiki.ros.org/rospy/Overview/Parameter%20Server
    """
    def __init__(self, name: str, out_topic: str, carrier: str = "tcp", persistent: bool = True, **kwargs):
        """
        The PropertiesPublisher using the ROS parameter server.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param persistent: bool: True if the parameter should be kept on closing node, False if it should be deleted or
                                    reset to its state before the node was started. Default is True
        """
        super().__init__(name, out_topic, carrier=carrier, **kwargs)
        self.persistent = persistent

        self.previous_property = False

    def establish(self, **kwargs):
        """
        Store the original property value in case it needs to be reset.
        """
        self.previous_property = rospy.get_param(self.out_topic, False)

    def publish(self, obj):
        """
        Publish the property to the middleware (parameter server).

        :param obj: object: Property to publish. If dict, will be set as a namespace
        """
        rospy.set_param(self.out_topic, obj)

    def close(self):
        """
        Close the publisher and reset the property to its original value if not persistent.
        """
        if hasattr(self, "out_topic") and not self.persistent:
            rospy.delete_param(self.out_topic)
            if self.previous_property:
                rospy.set_param(self.out_topic, self.previous_property)

    def __del__(self):
        self.close()


@Publishers.register("ROSMessage", "ros")
class ROSMessagePublisher(ROSPublisher):

    def __init__(self, name: str, out_topic: str, carrier: str = "tcp",
                 should_wait: bool = True, queue_size: int = QUEUE_SIZE, **kwargs):
        """
        The ROSMessagePublisher using the ROS message type inferred from the message type. Supports standard ROS msgs.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for PUB/SUB pattern. Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param queue_size: int: Queue size for the publisher. Default is 5
        """
        super().__init__(name, out_topic, carrier=carrier, should_wait=should_wait, queue_size=queue_size, **kwargs)

        self._publisher = None

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, obj=None, **kwargs):
        """
        Establish the connection and import the message requirements.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :param obj: object: Object to establish the connection to
        :return: bool: True if connection established, False otherwise
        """
        if obj is None:
            return
        obj_type = obj._type.split("/")
        import_msg = importlib.import_module(f"{obj_type[0]}.msg")
        msg_type = getattr(import_msg, obj_type[1])
        self._publisher = rospy.Publisher(self.out_topic, msg_type, queue_size=self.queue_size)
        established = self.await_connection(self._publisher, repeats=repeats)
        return self.check_establishment(established)

    def publish(self, obj):
        """
        Publish the object to the middleware.

        :param obj: object: ROS message to publish
        """
        if not self.established:
            established = self.establish(obj=obj)
            if not established:
                return
            else:
                time.sleep(0.2)
        self._publisher.publish(obj)



"""
Mirroring Example using Wrapyfi.

This script demonstrates the capability to mirror messages using the MiddlewareCommunicator within the Wrapyfi library.
The communication follows the PUB/SUB and REQ/REP patterns, allowing message publishing, listening, requesting, and replying
functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a Python object with user input and predefined message
    - Applying the PUB/SUB and REQ/REP patterns with mirroring

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)

Run:
    # Alternative 1: PUB/SUB
        # On machine 1 (or process 1): Publishing mode - Publisher waits for keyboard input and transmits message

        ``python3 mirroring_example.py --mode publish``

        # On machine 2 (or process 2): Listening mode - Listener waits for message and prints the received object

        ``python3 mirroring_example.py --mode listen``

    # Alternative 2: REQ/REP
        # On machine 1 (or process 1): Replying mode - Replier waits for a message, sends a reply, and prints the received object

        ``python3 mirroring_example.py --mode reply``

        # On machine 2 (or process 2): Requesting mode - Requester sends a predefined message and waits for a reply

        ``python3 mirroring_example.py --mode request``
"""

import argparse

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class MirrorCls(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        'NativeObject', '$mware', 'MirrorCls',
        '/example/read_msg',
        carrier='tcp', should_wait='$blocking')
    def read_msg(self, mware=None, msg='', blocking=True):
        """Exchange messages and mirror user input."""
        msg_ip = input('Type message: ')
        obj = {'msg': msg, 'msg_ip': msg_ip}
        return obj,

def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Mirroring Example using Wrapyfi.")
    parser.add_argument(
        "--mode", type=str, default="listen",
        choices={"publish", "listen", "request", "reply"},
        help="The communication mode (publish, listen, request, reply)"
    )
    parser.add_argument(
        "--mware", type=str, default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission"
    )
    return parser.parse_args()

def main(args):
    """Main function to initiate MirrorCls class and communication."""
    mirror = MirrorCls()
    mirror.activate_communication(MirrorCls.read_msg, mode=args.mode)

    while True:
        msg_object, = mirror.read_msg(mware=args.mware, msg=f"This argument message was sent by the {args.mode} script")
        if msg_object is not None:
            print(msg_object)


if __name__ == "__main__":
    args = parse_args()
    main(args)



import logging

from wrapyfi.connect.clients import Client, Clients


@Clients.register("MMO", "fallback")
class FallbackClient(Client):

    def __init__(self, name: str, in_topic: str, carrier: str = "", missing_middleware_object: str = "", **kwargs):
        logging.warning(f"Fallback client employed due to missing middleware or object type: "
                        f"{missing_middleware_object}")
        Client.__init__(self, name, in_topic, carrier=carrier, **kwargs)
        self.missing_middleware_object = missing_middleware_object

    def establish(self, repeats: int = -1, **kwargs):
        return None

    def request(self, *args, **kwargs):
        return None

    def _request(self, *args, **kwargs):
        return None

    def _await_reply(self):
        return None

    def close(self):
        return None


import logging
import json
import time
import base64
import io
from typing import Optional, Literal, Tuple

import numpy as np
import cv2
import yarp

from wrapyfi.connect.publishers import Publisher, Publishers, PublisherWatchDog
from wrapyfi.middlewares.yarp import YarpMiddleware
from wrapyfi.encoders import JsonEncoder


WATCHDOG_POLL_REPEAT = None


class YarpPublisher(Publisher):

    def __init__(self, name: str, out_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp", should_wait: bool = True,
                 persistent: bool = True, out_topic_connect: Optional[str] = None, yarp_kwargs: Optional[dict] = None, **kwargs):
        """
        Initialize the publisher.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param persistent: bool: Whether the publisher port should remain connected after closure. Default is True
        :param out_topic_connect: str: Name of the output topic connection alias '/' (e.g. '/topic:out') to connect to.
                                        None appends ':out' to the out_topic. Default is None
        :param yarp_kwargs: dict: Additional kwargs for  the Yarp middleware
        :param kwargs: dict: Additional kwargs for the publisher
        """
        super().__init__(name, out_topic, carrier=carrier, should_wait=should_wait, **kwargs)
        YarpMiddleware.activate(**yarp_kwargs or {})

        self.style = yarp.ContactStyle()
        self.style.persistent = persistent
        self.style.carrier = self.carrier

        self.out_topic_connect = out_topic + ":out" if out_topic_connect is None else out_topic_connect

    def await_connection(self, port, out_topic: Optional[str] = None, repeats: Optional[int] = None):
        """
        Wait for at least one subscriber to connect to the publisher.

        :param port: yarp.Port: Port to await connection to
        :param out_topic: str: Name of the output topic
        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        connected = False
        if out_topic is None:
            out_topic = self.out_topic
        logging.info(f"[YARP] Waiting for output connection: {out_topic}")
        if repeats is None:
            if self.should_wait:
                repeats = -1
            else:
                repeats = 1
            while repeats > 0 or repeats <= -1:
                repeats -= 1
                connected = port.getOutputCount() > 0
                if connected:
                    break
                time.sleep(0.02)
        logging.info(f"[YARP] Output connection established: {out_topic}")
        return connected

    def close(self):
        """
        Close the publisher
        """
        if hasattr(self, "_port") and self._port:
            if self._port is not None:
                self._port.close()

    def __del__(self):
        self.close()


@Publishers.register("NativeObject", "yarp")
class YarpNativeObjectPublisher(YarpPublisher):

    def __init__(self, name: str, out_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp", should_wait: bool = True,
                 persistent: bool = True, out_topic_connect: str = None, serializer_kwargs: Optional[dict] = None, **kwargs):
        """
        The NativeObject publisher using the BufferedPortBottle string construct assuming a combination of python native objects
        and numpy arrays as input. Serializes the data (including plugins) using the encoder and sends it as a string.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param persistent: bool: Whether the publisher port should remain connected after closure. Default is True
        :param out_topic_connect: str: Name of the output topic connection alias '/' (e.g. '/topic:out') to connect to.
                                        None appends ':out' to the out_topic. Default is None
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        super().__init__(name, out_topic, carrier=carrier, should_wait=should_wait, persistent=persistent,
                         out_topic_connect=out_topic_connect, **kwargs)
        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

        self._port = self._netconnect = None

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        self._port = yarp.BufferedPortBottle()
        self._port.open(self.out_topic)
        if self.style.persistent:
            self._netconnect = yarp.Network.connect(self.out_topic, self.out_topic_connect, self.style)
        else:
            self._netconnect = yarp.Network.connect(self.out_topic, self.out_topic_connect, self.carrier)
        established = self.await_connection(self._port, repeats=repeats)
        return self.check_establishment(established)

    def publish(self, obj):
        """
        Publish the object to the middleware.

        :param obj: object: Object to publish
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)
        obj_str = json.dumps(obj, cls=self._plugin_encoder, **self._plugin_kwargs,
                             serializer_kwrags=self._serializer_kwargs)
        obj_port = self._port.prepare()
        obj_port.clear()
        obj_port.addString(obj_str)
        self._port.write()


@Publishers.register("Image", "yarp")
class YarpImagePublisher(YarpPublisher):

    def __init__(self, name: str, out_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp", should_wait: bool = True,
                 persistent: bool = True, out_topic_connect: Optional[str] = None, width: int = -1, height: int = -1,
                 rgb: bool = True, fp: bool = False, jpg: bool = False, **kwargs):
        """
        The Image publisher using the BufferedPortImage construct assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param persistent: bool: Whether the publisher port should remain connected after closure. Default is True
        :param out_topic_connect: str: Name of the output topic connection alias '/' (e.g. '/topic:out') to connect to.
                                        None appends ':out' to the out_topic. Default is None
        :param width: int: Width of the image. Default is -1 meaning the width of the input image
        :param height: int: Height of the image. Default is -1 meaning the height of the input image
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be compressed as JPG. Default is False
        """
        super().__init__(name, out_topic, carrier=carrier, should_wait=should_wait, persistent=persistent,
                         out_topic_connect=out_topic_connect, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        self._port = self._type = self._netconnect = None

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        if self.jpg:
            self._port = yarp.BufferedPortBottle()
        elif self.rgb:
            self._port = yarp.BufferedPortImageRgbFloat() if self.fp else yarp.BufferedPortImageRgb()
        else:
            self._port = yarp.BufferedPortImageFloat() if self.fp else yarp.BufferedPortImageMono()
        self._type = np.float32 if self.fp else np.uint8
        self._port.open(self.out_topic)
        if self.style.persistent:
            self._netconnect = yarp.Network.connect(self.out_topic, self.out_topic_connect, self.style)
        else:
            self._netconnect = yarp.Network.connect(self.out_topic, self.out_topic_connect, self.carrier)
        established = self.await_connection(self._port, repeats=repeats)
        return self.check_establishment(established)

    def publish(self, img: np.ndarray):
        """
        Publish the image to the middleware.

        :param img: np.ndarray: Image to publish formatted as a cv2 image np.ndarray[img_height, img_width, channels] to publish
        """
        if img is None:
            return

        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)

        if 0 < self.width != img.shape[1] or 0 < self.height != img.shape[0] or \
                not ((img.ndim == 2 and not self.rgb) or (img.ndim == 3 and self.rgb and img.shape[2] == 3)):
            raise ValueError("Incorrect image shape for publisher")
        img = np.require(img, dtype=self._type, requirements='C')

        if self.jpg:
            img_str = np.array(cv2.imencode('.jpg', img)[1]).tostring()
            with io.BytesIO() as memfile:
                np.save(memfile, img_str)
                img_str = base64.b64encode(memfile.getvalue()).decode('ascii')
            img_port = self._port.prepare()
            img_port.clear()
            img_port.addString(img_str)
            self._port.write()

        else:
            img_port = self._port.prepare()
            img_port.resize(img.shape[1], img.shape[0])
            img_port.setExternal(img.data, img.shape[1], img.shape[0])
            self._port.write()


@Publishers.register("AudioChunk", "yarp")
class YarpAudioChunkPublisher(YarpPublisher):

    def __init__(self, name: str, out_topic: str, carrier: Literal["tcp", "udp", "mcast"] = "tcp", should_wait: bool = True,
                 persistent: bool = True, out_topic_connect: Optional[str] = None,
                 channels: int = 1, rate: int = 44100, chunk: int = -1, **kwargs):
        """
         The AudioChunk publisher using the Sound construct assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol (e.g. 'tcp'). Default is 'tcp'
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param persistent: bool: Whether the publisher port should remain connected after closure. Default is True
        :param out_topic_connect: str: Name of the output topic connection alias '/' (e.g. '/topic:out') to connect to.
                                        None appends ':out' to the out_topic. Default is None
        :param channels: int: Number of channels. Default is 1
        :param rate: int: Sampling rate. Default is 44100
        :param chunk: int: Chunk size. Default is -1 meaning that the chunk size is not fixed
        """
        super().__init__(name, out_topic, carrier=carrier, should_wait=should_wait, out_topic_connect=out_topic_connect,
                         persistent=persistent, **kwargs)
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

        self._sound_msg = self._port = self._netconnect = None

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        # create a dummy sound object for transmitting the sound props. This could be cleaner but left for future impl.
        self._port = yarp.Port()
        self._port.open(self.out_topic)
        self._netconnect = yarp.Network.connect(self.out_topic, self.out_topic_connect, self.carrier)
        self._sound_msg = yarp.Sound()
        self._sound_msg.setFrequency(self.rate)
        self._sound_msg.resize(self.chunk, self.channels)
        established = self.await_connection(self._port, out_topic=self.out_topic)
        if established:
            self._port.write(self._sound_msg)

        return self.check_establishment(established)

    def publish(self, aud: Tuple[np.ndarray, int]):
        """
        Publish the audio chunk to the middleware.

        :param aud: Tuple[np.ndarray, int]: Audio chunk to publish formatted as ((chunk_size, channels), samplerate)
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)

        aud, rate = aud
        if aud is None:
            return
        if 0 < self.rate != rate:
            raise ValueError("Incorrect audio rate for publisher")
        chunk, channels = aud.shape if len(aud.shape) > 1 else (aud.shape[0], 1)
        self.chunk = chunk if self.chunk == -1 else self.chunk
        self.channels = channels if self.channels == -1 else self.channels
        if 0 < self.chunk != chunk or 0 < self.channels != channels:
            raise ValueError("Incorrect audio shape for publisher")
        aud = np.require(aud, dtype=np.float32, requirements='C')

        for i in range(aud.size):
            self._sound_msg.set(int(aud.data[i] * 32767), i)  # Convert float samples to 16-bit int

        self._port.write(self._sound_msg)


@Publishers.register("Properties", "yarp")
class YarpPropertiesPublisher(YarpPublisher):

    def __init__(self, name, out_topic, **kwargs):
        super().__init__(name, out_topic, **kwargs)
        raise NotImplementedError



"""
A message publisher and listener for PIL (Pillow) images.

This script demonstrates the capability to transmit native Python objects and Pillow images in different formats using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern
allowing message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a nested dummy Python object with native objects and PIL (Pillow) images
    - Applying the PUB/SUB pattern with mirroring

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - NumPy: Used for creating image arrays (installed with Wrapyfi)
    - PIL (Pillow): Used for handling image objects

    Install using pip:
        ``pip install pillow``

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard and transmits message
        ``python3 pillow_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for message and prints the entire dummy object
        ``python3 pillow_example.py --mode listen``
"""

import argparse

try:
    import numpy as np
    from PIL import Image
except ImportError:
    print("Install PIL (Pillow) and NumPy before running this script.")

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class Notifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject", "$mware", "Notify", "/notify/test_native_exchange",
        carrier="", should_wait=True
    )
    def exchange_object(self, mware=None):
        msg = input("Type your message: ")
        imarray = np.random.rand(100, 100, 3) * 255
        ret = {
            "message": msg,
            "pillow_random": Image.fromarray(imarray.astype('uint8')).convert('RGBA'),
            "pillow_png": Image.open("../../assets/wrapyfi.png"),
            "pillow_jpg": Image.open("../../assets/wrapyfi.jpg")
        }
        return ret,


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--mode", type=str, default="publish",
        choices={"publish", "listen"},
        help="The transmission mode"
    )
    parser.add_argument(
        "--mware", type=str, default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission"
    )
    return parser.parse_args()


def main(args):
    """Main function to initiate Notify class and communication."""
    notifier = Notifier()
    notifier.activate_communication(Notifier.exchange_object, mode=args.mode)

    while True:
        msg_object, = notifier.exchange_object(mware=args.mware)
        print("Method result:", msg_object)


if __name__ == "__main__":
    args = parse_args()
    main(args)



"""
A message publisher and listener for ROS properties using Wrapyfi.

This script demonstrates the capability to transmit ROS properties, specifically
using the Properties message, using the MiddlewareCommunicator within the
Wrapyfi library. The communication follows the PUB/SUB pattern allowing
property publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the Properties message
    - Transmitting various data types as properties, such as strings, dictionaries, and non-persistent properties
    - Applying the PUB/SUB pattern with channeling and mirroring

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - ROS, ROSPy: Used for handling and creating ROS properties (refer to the Wrapyfi documentation for installation instructions)

    Ensure ROS is installed and the required message types are available.

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard input and transmits properties
        ``python3 ros_parameter_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for properties and prints the received properties
        ``python3 ros_parameter_example.py --mode listen``
"""

import argparse

from wrapyfi.connect.wrapper import MiddlewareCommunicator


class Notifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "Properties", "ros", "Notifier", "/notify/test_property_exchange",
        should_wait=True
    )
    @MiddlewareCommunicator.register(
        "Properties", "ros", "Notifier", "/notify/test_property_exchange/a",
        should_wait=True
    )
    @MiddlewareCommunicator.register(
        "Properties", "ros", "Notifier", "/notify/test_property_exchange/e",
        persistent=False, should_wait=True
    )
    def set_property(self):
        """Exchange ROS properties over ROS."""
        ret_str = input("Type your message: ")
        ret_multiprops = {"b": [1,2,3,4], "c": False, "d": 12.3}
        ret_non_persistent = "Non-persistent property which should be deleted on closure"
        return ret_multiprops, ret_str, ret_non_persistent


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="A message publisher and listener for ROS properties using Wrapyfi.")
    parser.add_argument(
        "--mode", type=str, default="publish",
        choices={"publish", "listen"},
        help="The transmission mode"
    )
    return parser.parse_args()


def main(args):
    """Main function to initiate Notify class and communication."""
    ros_message = Notifier()
    ros_message.activate_communication(Notifier.set_property, mode=args.mode)

    while True:
        my_dict_message, my_string_message, my_nonpersistent_message = ros_message.set_property()
        print("Method result:", my_string_message, my_dict_message, my_nonpersistent_message)


if __name__ == "__main__":
    args = parse_args()
    main(args)



"""
A message publisher and listener for native Python objects and PyArrow arrays.

This script demonstrates the capability to transmit native Python objects and PyArrow arrays using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern
allowing message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a nested dummy Python object with native objects and PyArrow arrays
    - Applying the PUB/SUB pattern with mirroring

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - PyArrow: Used for handling Apache Arrow arrays

    Install using pip:
        ``pip install pyarrow``

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard input and transmits message
        ``python3 pyarrow_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for message and prints the entire dummy object
        ``python3 pyarrow_example.py --mode listen``
"""

import argparse
try:
    import pyarrow as pa
except ImportError:
    print("Install PyArrow before running this script.")

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class Notifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject", "$mware", "Notifier", "/notify/test_arrow_exchange",
        carrier="tcp", should_wait=True
    )
    def exchange_object(self, mware=None):
        """Exchange messages with PyArrow arrays and other native Python objects."""
        msg = input("Type your message: ")
        ret = {
            "message": msg,
            "pyarrow_array": pa.array(range(100)),
        }
        return ret,


def parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        description="A message publisher and listener for native Python objects and PyArrow arrays.")
    parser.add_argument(
        "--mode", type=str, default="publish",
        choices={"publish", "listen"},
        help="The transmission mode"
    )
    parser.add_argument(
        "--mware", type=str, default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission"
    )
    return parser.parse_args()


def main(args):
    """Main function to initiate Notifier class and communication."""
    notifier = Notifier()
    notifier.activate_communication(Notifier.exchange_object, mode=args.mode)

    while True:
        msg_object, = notifier.exchange_object(mware=args.mware)
        print("Method result:", msg_object)


if __name__ == "__main__":
    args = parse_args()
    main(args)


import logging
import sys
import json
import queue
from typing import Optional, Any

import numpy as np
import rospy
import std_msgs.msg

from wrapyfi.connect.clients import Client, Clients
from wrapyfi.middlewares.ros import ROSMiddleware, ROSNativeObjectService, ROSImageService
from wrapyfi.encoders import JsonEncoder, JsonDecodeHook


class ROSClient(Client):
    def __init__(self, name: str, in_topic: str, carrier: str = "tcp",
                 ros_kwargs: Optional[dict] = None, **kwargs):
        """
        Initialize the client.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for rep/req pattern. Default is 'tcp'
        :param ros_kwargs: dict: Additional kwargs for the ROS middleware
        :param kwargs: dict: Additional kwargs for the client
        """
        if carrier or carrier != "tcp":
            logging.warning("[ROS] ROS does not support other carriers than TCP for REQ/REP pattern. Using TCP.")
            carrier = "tcp"
        super().__init__(name, in_topic, carrier=carrier, **kwargs)
        ROSMiddleware.activate(**ros_kwargs or {})

    def close(self):
        """
        Close the client.
        """
        if hasattr(self, "_client") and self._client:
            self._client.close()

    def __del__(self):
        self.close()


@Clients.register("NativeObject", "ros")
class ROSNativeObjectClient(ROSClient):

    def __init__(self, name: str, in_topic: str, carrier: str = "tcp", persistent: bool = True,
                 serializer_kwargs: Optional[dict] = None,
                 deserializer_kwargs: Optional[dict] = None, **kwargs):
        """
        The NativeObject client using the ROS String message assuming the data is serialized as a JSON string.
        Deserializes the data (including plugins) using the decoder and parses it to a Python object.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for rep/req pattern. Default is 'tcp'
        :param persistent: bool: Whether to keep the service connection alive across multiple service calls. Default is True
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, in_topic, carrier=carrier, **kwargs)
        self._client = None
        self._queue = queue.Queue(maxsize=1)

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}

        self.persistent = persistent

    def establish(self):
        """
        Establish the client's connection to the ROS service.
        """
        rospy.wait_for_service(self.in_topic)
        self._client = rospy.ServiceProxy(self.in_topic, ROSNativeObjectService, persistent=self.persistent)
        if self.persistent:
            self.established = True

    def request(self, *args, **kwargs) -> Any:
        """
        Send a request to the ROS service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        :return: Any: The response from the ROS service
        """
        if not self.established:
            self.establish()
        try:
            self._request(*args, **kwargs)
        except rospy.ServiceException as e:
            logging.error("[ROS] Service call failed: %s" % e)
        return self._await_reply()

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the ROS service.

        :param args: tuple: Positional arguments to send in the request.
        :param kwargs: dict: Keyword arguments to send in the request.
        """
        args_str = json.dumps([args, kwargs], cls=self._plugin_encoder, **self._plugin_kwargs,
                              serializer_kwrags=self._serializer_kwargs)
        args_msg = std_msgs.msg.String()
        args_msg.data = args_str

        msg = self._client(args_msg)
        obj = json.loads(msg.data, object_hook=self._plugin_decoder_hook, **self._deserializer_kwargs)
        self._queue.put(obj, block=False)

    def _await_reply(self) -> Any:
        """
        Wait for and return the reply from the ROS service.

        :return: Any: The response from the ROS service
        """
        try:
            reply = self._queue.get(block=True)
            return reply
        except queue.Full:
            logging.warning(f"[ROS] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__class__.__name__}")
            return None


@Clients.register("Image", "ros")
class ROSImageClient(ROSClient):

    def __init__(self, name: str, in_topic: str, carrier: str = "tcp", width: int = -1, height: int = -1, persistent: bool = True,
                 rgb: bool = True, fp: bool = False, serializer_kwargs: Optional[dict] = None, **kwargs):
        """
        The Image client using the ROS Image message parsed to a numpy array.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for rep/req pattern. Default is 'tcp'
        :param width: int: The width of the image. Default is -1
        :param height: int: The height of the image. Default is -1
        :param rgb: bool: Whether the image is RGB. Default is True
        :param fp: bool: Whether to utilize floating-point precision. Default is False
        :param persistent: bool: Whether to keep the service connection alive across multiple service calls. Default is True
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        if "jpg" in kwargs:
            logging.warning("[ROS] ROS currently does not support JPG encoding in REQ/REP. Using raw image.")
            kwargs.pop("jpg")
        super().__init__(name, in_topic, carrier=carrier, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp

        if self.fp:
            self._encoding = '32FC3' if self.rgb else '32FC1'
            self._type = np.float32
        else:
            self._encoding = 'bgr8' if self.rgb else 'mono8'
            self._type = np.uint8
        self._pixel_bytes = (3 if self.rgb else 1) * np.dtype(self._type).itemsize

        self._client = None
        self._queue = queue.Queue(maxsize=1)

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

        self.persistent = persistent

    def establish(self):
        """
        Establish the client's connection to the ROS service.
        """
        rospy.wait_for_service(self.in_topic)
        self._client = rospy.ServiceProxy(self.in_topic, ROSImageService, persistent=self.persistent)
        if self.persistent:
            self.established = True

    def request(self, *args, **kwargs):
        """
        Send a request to the ROS service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        :return: np.array: The received image from the ROS service
        """
        if not self.established:
            self.establish()
        try:
            self._request(*args, **kwargs)
        except rospy.ServiceException as e:
            logging.error("[ROS] Service call failed: %s" % e)
        return self._await_reply()

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the ROS service

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        """
        args_str = json.dumps([args, kwargs], cls=self._plugin_encoder, **self._plugin_kwargs,
                              serializer_kwrags=self._serializer_kwargs)
        args_msg = std_msgs.msg.String()
        args_msg.data = args_str
        msg = self._client(args_msg)
        self._queue.put((msg.height, msg.width, msg.encoding, msg.is_bigendian, msg.data), block=False)

    def _await_reply(self):
        """
        Wait for and return the reply from the ROS service.

        :return: np.array: The received image from the ROS service
        """
        try:
            height, width, encoding, is_bigendian, data = self._queue.get(block=True)
            if encoding != self._encoding:
                raise ValueError("Incorrect encoding for listener")
            if 0 < self.width != width or 0 < self.height != height or len(data) != height * width * self._pixel_bytes:
                raise ValueError("Incorrect image shape for listener")
            img = np.frombuffer(data, dtype=np.dtype(self._type).newbyteorder('>' if is_bigendian else '<')).reshape((height, width, -1))
            if img.shape[2] == 1:
                img = img.squeeze(axis=2)
            return img
        except queue.Full:
            logging.warning(f"[ROS] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__name__}")
            return None


@Clients.register("AudioChunk", "ros")
class ROSAudioChunkClient(ROSClient):

    def __init__(self, name: str, in_topic: str, carrier: str = "tcp", persistent: bool = True,
                 channels: int = 1, rate: int = 44100, chunk: int = -1, serializer_kwargs: Optional[dict] = None, **kwargs):
        """
        The AudioChunk client using the ROS Audio message parsed to a numpy array.

        :param name: str: Name of the client
        :param in_topic: str: Name of the input topic preceded by '/' (e.g. '/topic')
        :param carrier: str: Carrier protocol. ROS currently only supports TCP for rep/req pattern. Default is 'tcp'
        :param channels: int: Number of audio channels. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: The size of audio chunks. Default is -1
        :param persistent: bool: Whether to keep the service connection alive across multiple service calls. Default is True
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        """
        super().__init__(name, in_topic, carrier=carrier, **kwargs)
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

        self._client = self._req_msg = None
        self._queue = queue.Queue(maxsize=1)

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

        self.persistent = persistent

    def establish(self):
        """
        Establish the client's connection to the ROS service.
        """
        try:
            from wrapyfi_ros_interfaces.srv import ROSAudioService
        except ImportError:
            import wrapyfi
            logging.error("[ROS] Could not import ROSAudioService. "
                          "Make sure the ROS services in wrapyfi_extensions/wrapyfi_ros_interfaces are compiled. "
                          "Refer to the documentation for more information: \n" +
                          wrapyfi.__doc__ + "ros_interfaces_lnk.html")
            sys.exit(1)
        self._client = rospy.ServiceProxy(self.in_topic, ROSAudioService, persistent=self.persistent)
        self._req_msg = ROSAudioService._request_class()
        if self.persistent:
            self.established = True

    def request(self, *args, **kwargs):
        """
        Send a request to the ROS service

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        :return: Tuple[np.array, int]: The received audio chunk and rate from the ROS service
        """
        if not self.established:
            self.establish()
        try:
            self._request(*args, **kwargs)
        except rospy.ServiceException as e:
            logging.error("[ROS] Service call failed: %s" % e)
        return self._await_reply()

    def _request(self, *args, **kwargs):
        """
        Internal method to send a request to the ROS service.

        :param args: tuple: Positional arguments to send in the request
        :param kwargs: dict: Keyword arguments to send in the request
        """
        args_str = json.dumps([args, kwargs], cls=self._plugin_encoder, **self._plugin_kwargs,
                              serializer_kwrags=self._serializer_kwargs)
        args_msg = self._req_msg
        args_msg.request = args_str
        msg = self._client(args_msg).response
        self._queue.put((msg.chunk_size, msg.channels, msg.sample_rate, msg.encoding, msg.is_bigendian, msg.data), block=False)

    def _await_reply(self):
        """
        Wait for and return the reply from the ROS service.

        :return: Tuple[np.array, int]: The received audio chunk and rate from the ROS service
        """
        try:
            chunk, channels, rate, encoding, is_bigendian, data = self._queue.get(block=True)
            if 0 < self.rate != rate:
                raise ValueError("Incorrect audio rate for client")
            if encoding not in ['S16LE', 'S16BE']:
                raise ValueError("Incorrect encoding for client")
            if 0 < self.chunk != chunk or self.channels != channels or len(data) != chunk * channels * 4:
                raise ValueError("Incorrect audio shape for client")
            aud = np.frombuffer(data, dtype=np.dtype(np.float32).newbyteorder('>' if is_bigendian else '<')).reshape(
                (chunk, channels))
            # aud = aud / 32767.0
            return aud, rate
        except queue.Full:
            logging.warning(f"[ROS] Discarding data because queue is full. "
                            f"This happened due to bad synchronization in {self.__name__}")
            return None, self.rate




import logging
import atexit
import threading
import multiprocessing
import time
from collections import defaultdict
import json
from typing import Optional

import zmq

from wrapyfi.utils import SingletonOptimized
from wrapyfi.connect.wrapper import MiddlewareCommunicator

ZEROMQ_POST_OPTS = ["SUBSCRIBE", "UNSUBSCRIBE", "LINGER", "ROUTER_HANDOVER", "ROUTER_MANDATORY", "PROBE_ROUTER",
                    "XPUB_VERBOSE", "XPUB_VERBOSER", "REQ_CORRELATE", "REQ_RELAXED", "SNDHWM", "RCVHWM"]


class ZeroMQMiddlewarePubSub(metaclass=SingletonOptimized):
    """
    ZeroMQ PUB/SUB middleware wrapper. This class is a singleton, so it can be instantiated only once. The ``activate``
    method should be called to initialize the middleware. The ``deinit`` method should be called to deinitialize the
    middleware and destroy all connections. The ``activate`` and ``deinit`` methods are automatically called when the
    class is instantiated and when the program exits, respectively.
    """
    class ZeroMQSharedMonitorData:
        """
        Shared data class for the ZeroMQ PUB/SUB monitor. This class is used to share data between the main process and
        the monitor listener process/thread.
        """
        def __init__(self, use_multiprocessing: bool = False):
            """
            Initialize the shared data class.

            :param use_multiprocessing: bool: Whether to use multiprocessing or threading
            """
            self.use_multiprocessing = use_multiprocessing
            if use_multiprocessing:
                manager = multiprocessing.Manager()
                self.shared_topics = manager.list()
                self.shared_connections = manager.dict()
                self.lock = manager.Lock()
            else:
                self.shared_topics = []
                self.shared_connections = {}
                self.lock = threading.Lock()

        def add_topic(self, topic: str):
            """
            Add a topic to the shared data class.

            :param topic: str: The topic to add
            """
            with self.lock:
                self.shared_topics.append(topic)

        def remove_topic(self, topic: str):
            """
            Remove a topic from the shared data class.

            :param topic: str: The topic to remove
            """
            with self.lock:
                if topic in self.shared_topics:
                    self.shared_topics.remove(topic)

        def get_topics(self):
            """
            Get the list of topics in the shared data class.

            :return: list: The list of topics
            """
            with self.lock:
                return list(self.shared_topics)

        def update_connection(self, topic: str, data: dict):
            """
            Update the connection data for a topic, e.g. the number of subscribers.

            :param topic: str: The topic to update
            :param data: dict: The connection data
            """
            with self.lock:
                self.shared_connections[topic] = data

        def remove_connection(self, topic: str):
            """
            Remove the connection data for a topic.

            :param topic: str: The topic to remove
            """
            with self.lock:
                if topic in list(self.shared_connections.keys()):
                    del self.shared_connections[topic]

        def get_connections(self):
            """
            Get the connection data for all topics.

            :return: dict: The connection data for all topics
            """
            with self.lock:
                return dict(self.shared_connections)

        def is_connected(self, topic: str):
            """
            Check whether a topic is connected.

            :param topic: str: The topic to check
            """
            with self.lock:
                if topic in list(self.shared_connections.keys()):
                    return True
                else:
                    return False

    @staticmethod
    def activate(**kwargs):
        """
        Activate the ZeroMQ PUB/SUB middleware. This method should be called to initialize the middleware.

        :param kwargs: dict: Keyword arguments to be passed to the ZeroMQ initialization function
        """
        zeromq_post_kwargs = {}
        zeromq_pre_kwargs = {}
        for key, value in kwargs.items():
            try:
                getattr(zmq, key)
                if key in ZEROMQ_POST_OPTS:
                    zeromq_post_kwargs[key] = value
                else:
                    zeromq_pre_kwargs[key] = value
            except AttributeError:
                pass

        ZeroMQMiddlewarePubSub(zeromq_proxy_kwargs=kwargs, zeromq_post_kwargs=zeromq_post_kwargs, **zeromq_pre_kwargs)

    def __init__(self, zeromq_proxy_kwargs: Optional[dict] = None, zeromq_post_kwargs: Optional[dict] = None, **kwargs):
        """
        Initialize the ZeroMQ PUB/SUB middleware. This method is automatically called when the class is instantiated.

        :param zeromq_proxy_kwargs: Optional[dict]: Keyword arguments to be passed to the ZeroMQ proxy initialization function
        :param zeromq_post_kwargs: Optional[dict]: Keyword arguments to be passed to the ZeroMQ initialization function (these are ZeroMQ options)
        :param kwargs: dict: Keyword arguments to be passed to the ZeroMQ initialization function
        """
        self.zeromq_proxy_kwargs = zeromq_proxy_kwargs or {}
        self.zeromq_kwargs = zeromq_post_kwargs or {}
        logging.info("Initialising ZeroMQ PUB/SUB middleware")
        self.ctx = zmq.Context.instance()
        for socket_property in kwargs.items():
            if isinstance(socket_property[1], str):
                self.ctx.setsockopt_string(getattr(zmq, socket_property[0]), socket_property[1])
            else:
                self.ctx.setsockopt(getattr(zmq, socket_property[0]), socket_property[1])
        atexit.register(MiddlewareCommunicator.close_all_instances)
        atexit.register(self.deinit)

        # start the pubsub proxy and monitor
        if zeromq_proxy_kwargs is not None and zeromq_proxy_kwargs:
            # start the pubsub monitor listener
            if zeromq_proxy_kwargs.get("pubsub_monitor_listener_spawn", False):
                if zeromq_proxy_kwargs["pubsub_monitor_listener_spawn"] == "process":
                    self.shared_monitor_data = self.ZeroMQSharedMonitorData(use_multiprocessing=True)
                    self.monitor = multiprocessing.Process(name='zeromq_pubsub_monitor_listener',
                                                           target=self.__init_monitor_listener,
                                                           kwargs=zeromq_proxy_kwargs)
                    self.monitor.daemon = True
                    self.monitor.start()
                else:  # if threaded
                    self.shared_monitor_data = self.ZeroMQSharedMonitorData(use_multiprocessing=False)
                    self.monitor = threading.Thread(name='pubsub_monitor_listener_spawn',
                                                    target=self.__init_monitor_listener, kwargs=zeromq_proxy_kwargs)
                    self.monitor.setDaemon(True)
                    self.monitor.start()

            time.sleep(1)

            if zeromq_proxy_kwargs.get("start_proxy_broker", False):
                if zeromq_proxy_kwargs["proxy_broker_spawn"] == "process":
                    self.proxy = multiprocessing.Process(name='zeromq_pubsub_broker', target=self.__init_proxy, kwargs=zeromq_proxy_kwargs)
                    self.proxy.daemon = True
                    self.proxy.start()
                else:  # if threaded
                    self.proxy = threading.Thread(name='zeromq_pubsub_broker', target=self.__init_proxy, kwargs=zeromq_proxy_kwargs)
                    self.proxy.setDaemon(True)
                    self.proxy.start()
            pass
                
    @staticmethod
    def proxy_thread(socket_pub_address: str = "tcp://127.0.0.1:5555",
                     socket_sub_address: str = "tcp://127.0.0.1:5556",
                     inproc_address: str = "inproc://monitor"):
        """
        Proxy thread for the ZeroMQ PUB/SUB proxy.

        :param socket_pub_address: str: The address of the PUB socket
        :param socket_sub_address: str: The address of the SUB socket
        :param inproc_address: str: The address of the inproc socket (connections within the same process, for exchanging subscription data between the proxy and the monitor)
        """
        context = zmq.Context.instance()
        xpub = context.socket(zmq.XPUB)
        xsub = context.socket(zmq.XSUB)
        xpub.setsockopt(zmq.XPUB_VERBOSE, 1)

        xpub.bind(socket_pub_address)
        xsub.bind(socket_sub_address)

        monitor = context.socket(zmq.PUB)
        monitor.bind(inproc_address)

        zmq.proxy(xpub, xsub, monitor)

    @staticmethod
    def subscription_monitor_thread(inproc_address: str = "inproc://monitor", socket_sub_address: str = "tcp://127.0.0.1:5556",
                                    pubsub_monitor_topic: str = "ZEROMQ/CONNECTIONS", verbose: bool = False):
        """
        Subscription monitor thread for the ZeroMQ PUB/SUB proxy.

        :param inproc_address: str: The address of the inproc socket (connections within the same process, for exchanging subscription data between the proxy and the monitor)
        :param socket_sub_address: str: The address of the SUB socket
        :param pubsub_monitor_topic: str: The topic to use for publishing subscription data
        :param verbose: bool: Whether to print debug messages
        """
        context = zmq.Context.instance()
        subscriber = context.socket(zmq.SUB)
        subscriber.connect(inproc_address)
        subscriber.setsockopt_string(zmq.SUBSCRIBE, "")

        # pub socket to publish subscriber counts.
        publisher = context.socket(zmq.PUB)
        publisher.connect(socket_sub_address)

        topic_subscriber_count = defaultdict(int)

        while True:
            time.sleep(1)
            try:
                message = subscriber.recv()
                if verbose:
                    logging.info(f"[ZeroMQ BROKER] Raw message: {message}")

                # ensure the message is a subscription/unsubscription message
                if len(message) > 1 and (message[0] == 1 or message[0] == 0):
                    event = message[0]
                    topic = message[1:].decode('utf-8')

                    if verbose:
                        logging.info(f"[ZeroMQ BROKER] Received event: {event}, topic: {topic}")

                    # avoid processing messages on the monitor topic.
                    if topic == pubsub_monitor_topic:
                        continue

                    # update the count of subscribers for the topic
                    if event == 1:  # subscribe
                        topic_subscriber_count[topic] += 1
                    elif event == 0:  # unsubscribe
                        topic_subscriber_count[topic] = 0

                    if verbose:
                        logging.info(f"[ZeroMQ BROKER] Current topic subscriber count: {dict(topic_subscriber_count)}")

                    # publish the updated counts
                    publisher.send_multipart(
                        [pubsub_monitor_topic.encode(), json.dumps(dict(topic_subscriber_count)).encode()]
                    )
            except Exception as e:
                logging.error(f"[ZeroMQ BROKER] An error occurred in the ZeroMQ subscription monitor: {str(e)}")

    def __init_proxy(self, socket_pub_address: str = "tcp://127.0.0.1:5555", socket_sub_address: str = "tcp://127.0.0.1:5556",
                     pubsub_monitor_topic: str = "ZEROMQ/CONNECTIONS",
                     **kwargs):
        """
        Initialize the ZeroMQ PUB/SUB proxy and subscription monitor.

        :param socket_pub_address: str: The address of the PUB socket
        :param socket_sub_address: str: The address of the SUB socket
        :param pubsub_monitor_topic: str: The topic to use for publishing subscription data
        :param kwargs: dict: Keyword arguments to be passed to the ZeroMQ initialization function
        """
        inproc_address = "inproc://monitor"

        threading.Thread(target=self.proxy_thread,
                         kwargs={"socket_pub_address": socket_pub_address,
                                 "socket_sub_address": socket_sub_address,
                                 "inproc_address": inproc_address}).start(),

        threading.Thread(target=self.subscription_monitor_thread,
                         kwargs={"socket_sub_address": socket_sub_address,
                                 "inproc_address": inproc_address,
                                 "pubsub_monitor_topic": pubsub_monitor_topic,
                                 "verbose": kwargs.get("verbose", False)}).start()

    def __init_monitor_listener(self, socket_pub_address: str = "tcp://127.0.0.1:5555", 
                                pubsub_monitor_topic: str = "ZEROMQ/CONNECTIONS",
                                verbose: bool = False, **kwargs):
        """
        Initialize the ZeroMQ PUB/SUB monitor listener.

        :param socket_pub_address: str: The address of the PUB socket
        :param pubsub_monitor_topic: str: The topic to use for publishing subscription data
        :param verbose: bool: Whether to print debug messages
        """
        try:
            context = zmq.Context()
            subscriber = context.socket(zmq.SUB)

            subscriber.connect(socket_pub_address)
            subscriber.setsockopt_string(zmq.SUBSCRIBE, pubsub_monitor_topic)

            while True:
                _, message = subscriber.recv_multipart()
                data = json.loads(message.decode('utf-8'))
                topic = list(data.keys())[0]
                if verbose:
                    logging.info(f"[ZeroMQ] Topic: {topic}, Data: {data}")

                if topic in self.shared_monitor_data.get_topics():
                    self.shared_monitor_data.update_connection(topic, data)
                    if data[topic] == 0:
                        logging.info(f"[ZeroMQ] Subscriber disconnected from topic: {topic}")
                        self.shared_monitor_data.remove_connection(topic)
                    else:
                        logging.info(f"[ZeroMQ] Subscriber connected to topic: {topic}")

                if verbose:
                    for monitored_topic in self.shared_monitor_data.get_topics():
                        logging.info(f"[ZeroMQ] Monitored topic from main process: {monitored_topic}")

        except Exception as e:
            logging.error(f"[ZeroMQ] An error occurred in the ZeroMQ subscription monitor listener: {str(e)}")

    @staticmethod
    def deinit():
        logging.info("Deinitializing ZeroMQ middleware")
        zmq.Context.instance().destroy()


class ZeroMQMiddlewareReqRep(metaclass=SingletonOptimized):
    """
    ZeroMQ REQ/REP middleware wrapper. This class is a singleton, so it can be instantiated only once. The ``activate``
    method should be called to initialize the middleware. The ``deinit`` method should be called to deinitialize the
    middleware and destroy all connections. The ``activate`` and ``deinit`` methods are automatically called when the
    class is instantiated and when the program exits, respectively.
    """
    @staticmethod
    def activate(**kwargs):
        """
        Activate the ZeroMQ REQ/REP middleware. This method should be called to initialize the middleware.

        :param kwargs: dict: Keyword arguments to be passed to the ZeroMQ initialization function
        """
        zeromq_post_kwargs = {}
        zeromq_pre_kwargs = {}
        for key, value in kwargs.items():
            try:
                getattr(zmq, key)
                if key in ZEROMQ_POST_OPTS:
                    zeromq_post_kwargs[key] = value
                else:
                    zeromq_pre_kwargs[key] = value
            except AttributeError:
                pass

        ZeroMQMiddlewareReqRep(zeromq_proxy_kwargs=kwargs, zeromq_post_kwargs=zeromq_post_kwargs, **zeromq_pre_kwargs)

    def __init__(self, zeromq_proxy_kwargs: Optional[dict] = None, zeromq_post_kwargs: Optional[dict] = None, *args, **kwargs):
        """
        Initialize the ZeroMQ REQ/REP middleware. This method is automatically called when the class is instantiated.

        :param zeromq_proxy_kwargs: Optional[dict]: Keyword arguments to be passed to the ZeroMQ proxy initialization function
        :param zeromq_post_kwargs: Optional[dict]: Keyword arguments to be passed to the ZeroMQ initialization function (these are ZeroMQ options)
        :param args: list: Positional arguments to be passed to the ZeroMQ initialization function
        :param kwargs: dict: Keyword arguments to be passed to the ZeroMQ initialization function
        """
        self.zeromq_proxy_kwargs = zeromq_proxy_kwargs or {}
        self.zeromq_kwargs = zeromq_post_kwargs or {}
        logging.info("Initialising ZeroMQ REQ/REP middleware")
        self.ctx = zmq.Context.instance()
        for socket_property in kwargs.items():
            if isinstance(socket_property[1], str):
                self.ctx.setsockopt_string(getattr(zmq, socket_property[0]), socket_property[1])
            else:
                self.ctx.setsockopt(getattr(zmq, socket_property[0]), socket_property[1])
        atexit.register(MiddlewareCommunicator.close_all_instances)
        atexit.register(self.deinit)

        if zeromq_proxy_kwargs is not None and zeromq_proxy_kwargs:
            if zeromq_proxy_kwargs["proxy_broker_spawn"] == "process":
                self.proxy = multiprocessing.Process(name='zeromq_reqrep_broker', target=self.__init_device, kwargs=zeromq_proxy_kwargs)
                self.proxy.daemon = True
                self.proxy.start()
            else:  # if threaded
                self.proxy = threading.Thread(name='zeromq_reqrep_broker', target=self.__init_device, kwargs=zeromq_proxy_kwargs)
                self.proxy.setDaemon(True)
                self.proxy.start()
            pass

    @staticmethod
    def __init_device(socket_rep_address: str = "tcp://127.0.0.1:5559", socket_req_address: str = "tcp://127.0.0.1:5560", **kwargs):
        """
        Initialize the ZeroMQ REQ/REP device broker.

        :param socket_rep_address: str: The address of the REP socket
        :param socket_req_address: str: The address of the REQ socket
        """
        xrep = zmq.Context.instance().socket(zmq.XREP)
        try:
            xrep.bind(socket_rep_address)
        except zmq.ZMQError as e:
            logging.error(f"[ZeroMQ] {e} {socket_rep_address}")
            return
        xreq = zmq.Context.instance().socket(zmq.XREQ)
        try:
            xreq.bind(socket_req_address)
        except zmq.ZMQError as e:
            logging.error(f"[ZeroMQ] {e} {socket_req_address}")
            return
        # logging.info(f"[ZeroMQ] Intialising REQ/REP device broker")
        zmq.proxy(xrep, xreq)

    @staticmethod
    def deinit():
        logging.info("Deinitializing ZeroMQ middleware")
        zmq.Context.instance().destroy()


class ZeroMQMiddlewareParamServer(metaclass=SingletonOptimized):
    """
    ZeroMQ parameter server middleware wrapper. This class is a singleton, so it can be instantiated only once. The
    ``activate`` method should be called to initialize the middleware. The ``deinit`` method should be called to
    deinitialize the middleware and destroy all connections. The ``activate`` and ``deinit`` methods are automatically
    called when the class is instantiated and when the program exits, respectively.

    Note: This parameter server is experimental and not fully tested.
    """
    @staticmethod
    def activate(**kwargs):
        zeromq_post_kwargs = {}
        zeromq_pre_kwargs = {}
        for key, value in kwargs.items():
            try:
                getattr(zmq, key)
                if key in ZEROMQ_POST_OPTS:
                    zeromq_post_kwargs[key] = value
                else:
                    zeromq_pre_kwargs[key] = value
            except AttributeError:
                pass

        ZeroMQMiddlewareParamServer(zeromq_proxy_kwargs=kwargs, zeromq_post_kwargs=zeromq_post_kwargs, **zeromq_pre_kwargs)

    def __init__(self, zeromq_proxy_kwargs: Optional[dict] = None, zeromq_post_kwargs: Optional = None, *args, **kwargs):
        """
        Initialize the ZeroMQ parameter server middleware. This method is automatically called when the class is
        instantiated.

        :param zeromq_proxy_kwargs: Optional[dict]: Keyword arguments to be passed to the ZeroMQ proxy initialization function
        :param zeromq_post_kwargs: Optional[dict]: Keyword arguments to be passed to the ZeroMQ initialization function (these are ZeroMQ options)
        :param kwargs: dict: Keyword arguments to be passed to the ZeroMQ initialization function
        """
        self.zeromq_proxy_kwargs = zeromq_proxy_kwargs or {}
        self.zeromq_kwargs = zeromq_post_kwargs or {}
        logging.info("Initialising ZeroMQ Parameter Server")
        self.ctx = zmq.Context.instance()
        for socket_property in kwargs.items():
            if isinstance(socket_property[1], str):
                self.ctx.setsockopt_string(getattr(zmq, socket_property[0]), socket_property[1])
            else:
                self.ctx.setsockopt(getattr(zmq, socket_property[0]), socket_property[1])

        atexit.register(MiddlewareCommunicator.close_all_instances)
        atexit.register(self.deinit)

        if zeromq_proxy_kwargs is not None and zeromq_proxy_kwargs:
            self.manager = multiprocessing.Manager()
            self.params = self.manager.dict()
            self.params["WRAPYFI_ACTIVE"] = "True"
            if zeromq_proxy_kwargs["proxy_broker_spawn"] == "process":
                self.param_broadcaster = multiprocessing.Process(name='zeromq_param_broadcaster', target=self.__init_broadcaster,
                                                            kwargs=zeromq_proxy_kwargs, args=(self.params,))
                self.param_broadcaster.daemon = True
                self.param_broadcaster.start()
                self.param_server = multiprocessing.Process(name='zeromq_param_server', target=self.__init_server,
                                                            kwargs=zeromq_proxy_kwargs, args=(self.params,))
                self.param_server.daemon = True
                self.param_server.start()
            else:  # if threaded

                self.param_broadcaster = threading.Thread(name='zeromq_param_broadcaster', target=self.__init_broadcaster,
                                                     kwargs=zeromq_proxy_kwargs, args=(self.params,))
                self.param_broadcaster.setDaemon(True)
                self.param_broadcaster.start()
                self.param_server = threading.Thread(name='zeromq_param_server', target=self.__init_server,
                                                     kwargs=zeromq_proxy_kwargs, args=(self.params,))
                self.param_server.setDaemon(True)
                self.param_server.start()
            pass

    @staticmethod
    def __init_broadcaster(params, param_pub_address: str = "tcp://127.0.0.1:5655", param_sub_address: str = "tcp://127.0.0.1:5656",
                           param_poll_interval=1, verbose=False, **kwargs):
        """
        Initialize the ZeroMQ parameter server broadcaster.

        :param params: dict: The parameters to be broadcasted
        :param param_pub_address: str: The address of the PUB socket
        :param param_sub_address: str: The address of the SUB socket
        :param param_poll_interval: int: The polling interval for the parameter server
        :param verbose: bool: Whether to print debug messages
        """
        update_trigger = False
        cached_params = {}
        root_topics = set()

        ctx = zmq.Context.instance()

        # create XPUB
        xpub_socket = ctx.socket(zmq.XPUB)
        xpub_socket.bind(param_pub_address)

        # create XSUB
        xsub_socket = ctx.socket(zmq.XSUB)
        xsub_socket.bind(param_sub_address)

        # connect a PUB socket to send parameters
        param_server = ctx.socket(zmq.PUB)
        param_server.connect(param_sub_address)

        poller = zmq.Poller()
        poller.register(xpub_socket, zmq.POLLIN)
        poller.register(xsub_socket, zmq.POLLIN)
        if verbose:
            logging.info(f"[ZeroMQ] Intialising PUB/SUB device broker")
        while True:
            # get event
            event = dict(poller.poll(param_poll_interval))
            if xpub_socket in event:
                message = xpub_socket.recv_multipart()
                if verbose:
                    logging.info("[ZeroMQ BROKER] xpub_socket recv message: %r" % message)
                if message[0].startswith(b"\x00"):
                    root_topics.remove(message[0][1:].decode("utf-8"))
                elif message[0].startswith(b"\x01"):
                    root_topics.add(message[0][1:].decode("utf-8"))
                xsub_socket.send_multipart(message)

            if xsub_socket in event:
                message = xsub_socket.recv_multipart()
                if verbose:
                    logging.info("[ZeroMQ BROKER] xsub_socket recv message: %r" % message)
                if message[0].startswith(b"\x01") or message[0].startswith(b"\x00"):
                    xpub_socket.send_multipart(message)
                else:
                    fltr_key = message[0].decode("utf-8")
                    fltr_message = {key: val for key, val in params.items()
                                    if key.startswith(fltr_key)}
                    if verbose:
                        logging.info("[ZeroMQ BROKER] xsub_socket filtered message: %r" % fltr_message)
                    for key, val in fltr_message.items():
                        prefix, param = key.rsplit("/", 1) if "/" in key else ("", key)
                        xpub_socket.send_multipart([prefix.encode("utf-8"), param.encode("utf-8"), val.encode("utf-8")])
                # xpub_socket.send_multipart(message)

            if event:
                update_trigger = True

            if param_server is not None:
                update_trigger, cached_params = ZeroMQMiddlewareParamServer.publish_params(
                    param_server, params, cached_params, root_topics, update_trigger)

    @staticmethod
    def publish_params(param_server, params: dict, cached_params: dict, root_topics: set, update_trigger: bool):
        """
        Publish parameters to the parameter server.

        :param param_server: zmq.Socket: The parameter server socket
        :param params: dict: The parameters to be published
        :param cached_params: dict: The cached parameters. This is used to check whether the parameters have changed
        :param root_topics: set: The root topics. This is used to check whether there are any active subscribers
        :param update_trigger: bool: Whether to trigger an update of the parameters
        :return: Tuple[bool, dict]: Whether to trigger an update of the parameters and the cached parameters
        """
        # check if there are any subscribed clients before publishing updates
        if not root_topics:
            return None, None

        if not any((update_trigger, root_topics)) and cached_params == params:
            return None, None
        else:
            time.sleep(0.01)
            update_trigger = False
            cached_params = params.copy()

        # publish updates for all parameters to subscribed clients
        for key, val in params.items():
            prefix, param = key.rsplit("/", 1) if "/" in key else ("", key)
            param_server.send_multipart([prefix.encode("utf-8"), param.encode("utf-8"), val.encode("utf-8")])

        return update_trigger, cached_params

    @staticmethod
    def __init_server(params: dict, param_reqrep_address: str = "tcp://127.0.0.1:5659", **kwargs):
        """
        Initialize the ZeroMQ parameter server.

        :param params: dict: The parameters to be published
        :param param_reqrep_address: str: The address of the REQ/REP socket
        """
        ctx = zmq.Context.instance()
        request_server = ctx.socket(zmq.REP)
        request_server.bind(param_reqrep_address)

        while True:
            request = request_server.recv_string()
            if request.startswith("get"):
                try:
                    # extract the parameter name and namespace prefix from the request
                    prefix, param = request[4:].rsplit("/", 1) if "/" in request[4:] else ("", request[4:])
                    # construct the full parameter name with the namespace prefix
                    full_param = "/".join([prefix, param]) if prefix else param
                    if full_param in params:
                        request_server.send_string(str(params[full_param]))
                    else:
                        request_server.send_string("error:::parameter does not exist")
                except ValueError:
                    request_server.send_string("error:::malformed request")
            elif request.startswith("read"):
                try:
                    # extract the parameter name and namespace prefix from the request
                    prefix = request[5:]
                    # construct the full parameter name with the namespace prefix
                    if any(param.startswith(prefix) for param in params.keys()):
                        request_server.send_string(f"success:::{prefix}")
                    else:
                        request_server.send_string("error:::parameter does not exist")
                except ValueError:
                    request_server.send_string("error:::malformed request")
            elif request.startswith("set"):
                try:
                    # extract the parameter name, namespace prefix and value from the request
                    prefix, param, value = request[4:].rsplit("/", 2)
                    # construct the full parameter name with the namespace prefix
                    full_param = "/".join([prefix, param]) if prefix else param
                    params[full_param] = value
                    request_server.send_string(f"success:::{prefix}")
                except ValueError:
                    request_server.send_string("error:::malformed request")
            elif request.startswith("delete"):
                try:
                    # extract the parameter name and namespace prefix from the request
                    prefix, param = request[7:].rsplit("/", 1)
                    # construct the full parameter name with the namespace prefix
                    full_param = "/".join([prefix, param]) if prefix else param
                    if full_param in params:
                        del params[full_param]
                        request_server.send_string(f"success:::{prefix}")
                    else:
                        request_server.send_string("error:::parameter does not exist")
                except ValueError:
                    request_server.send_string("error:::malformed request")
            else:
                request_server.send_string("error:::invalid request")

    @staticmethod
    def deinit():
        """
        Deinitialize the ZeroMQ parameter server.
        """
        logging.info("Deinitializing ZeroMQ Parameter Server")
        zmq.Context.instance().destroy()
